@@ -1,10 +1,6 @@
 #import <Foundation/Foundation.h>
 
-@interface LineBufferPosition : NSObject {
-    long long absolutePosition_;
-    int yOffset_;
-    BOOL extendsToEndOfLine_;
-}
+@interface LineBufferPosition : NSObject
 
 // Absolute position - bytes from start of line buffer plus total number of bytes that ever have
 // been dropped.
@@ -22,6 +18,8 @@
 + (LineBufferPosition *)position;
 - (LineBufferPosition *)predecessor;
 
+- (BOOL)isEqualToLineBufferPosition:(LineBufferPosition *)other;
+
 @end
 
 @interface LineBufferPositionRange : NSObject
