@@ -11,16 +11,18 @@
 #import "iTermRateLimitedUpdate.h"
 #import <MetalKit/MetalKit.h>
 
+@interface NSClipView(Private)
+- (BOOL)_shouldShowOverlayScrollersForScrollToPoint:(CGPoint)point;
+@end
+
 @implementation iTermMetalClipView {
-    iTermRateLimitedUpdate *_rateLimit;
+    NSInteger _disableShowingOverlayScrollers;
 }
 
 - (instancetype)initWithFrame:(NSRect)frameRect {
     self = [super initWithFrame:frameRect];
     if (self) {
         self.copiesOnScroll = NO;
-        _rateLimit = [[iTermRateLimitedUpdate alloc] init];
-        _rateLimit.minimumInterval = 1.0 / 60.0;
     }
     return self;
 }
@@ -28,14 +30,22 @@
 - (void)scrollToPoint:(NSPoint)newOrigin {
     [super scrollToPoint:newOrigin];
     if (_useMetal) {
-        BOOL performed = [_rateLimit tryPerformRateLimitedBlock:^{
-            [_metalView draw];
-        }];
-        if (!performed) {
-            [_metalView setNeedsDisplay:YES];
-        }
+        [_metalView setNeedsDisplay:YES];
     }
 }
 
+- (void)performBlockWithoutShowingOverlayScrollers:(void (^ NS_NOESCAPE)(void))block {
+    _disableShowingOverlayScrollers += 1;
+    block();
+    _disableShowingOverlayScrollers -= 1;
+}
+
+- (BOOL)_shouldShowOverlayScrollersForScrollToPoint:(CGPoint)point {
+    if (_disableShowingOverlayScrollers) {
+        return NO;
+    }
+    return [super _shouldShowOverlayScrollersForScrollToPoint:point];
+}
+
 @end
 
