@@ -25,9 +25,10 @@
 
 @interface NSFileManager (iTerm)
 
-- (NSString *)applicationSupportDirectory;
++ (NSString *)pathToSaveFileInFolder:(NSString *)destinationDirectory preferredName:(NSString *)preferredName;
 
 - (NSString *)legacyApplicationSupportDirectory;
+- (NSString *)applicationSupportDirectory;
 
 - (NSString *)temporaryDirectory;
 
@@ -35,6 +36,11 @@
 
 - (BOOL)directoryIsWritable:(NSString *)dir;
 
+// Returns YES if the file looks like it might be on a local filesystem, but doesn't check if it
+// actually exists.
+- (BOOL)fileIsLocal:(NSString *)filename
+additionalNetworkPaths:(NSArray<NSString *> *)additionalNetworkPaths;
+
 // Returns YES if the file exists on a local (non-network) filesystem.
 - (BOOL)fileExistsAtPathLocally:(NSString *)filename
          additionalNetworkPaths:(NSArray<NSString *> *)additionalNetworkpaths;
@@ -53,10 +59,19 @@
 - (NSString *)scriptsPath;
 
 // Path to special auto-launch script that is run at startup.
-- (NSString *)autolaunchScriptPath;
+- (NSString *)legacyAutolaunchScriptPath;  // applescript
+- (NSString *)autolaunchScriptPath;  // scripting API
 
 // Path to special file that, if it exists at launch time, suppresses autolaunch script and
 // window restoration.
 - (NSString *)quietFilePath;
+- (BOOL)directoryEmpty:(NSString *)path;
+- (BOOL)itemIsSymlink:(NSString *)path;
+- (BOOL)itemIsDirectory:(NSString *)path;
+- (NSArray<NSString *> *)it_itemsInDirectory:(NSString *)path;
+- (NSString *)libraryDirectoryFor:(NSString *)app;
+
+- (id)monitorFile:(NSString *)file block:(void (^)(long flags))block;
+- (void)stopMonitoringFileWithToken:(id)token;
 
 @end
