@@ -10,16 +10,20 @@
 #import "PseudoTerminal+Private.h"
 
 #import "DebugLogging.h"
+#import "NSArray+iTerm.h"
 #import "NSDictionary+iTerm.h"
 #import "NSImage+iTerm.h"
 #import "NSStringITerm.h"
 #import "iTermAdvancedSettingsModel.h"
 #import "iTermApplication.h"
 #import "iTermColorPresets.h"
-#import "iTermKeyBindingMgr.h"
+#import "iTermKeyBindingAction.h"
+#import "iTermKeystroke.h"
 #import "iTermRootTerminalView.h"
 #import "iTermSystemVersion.h"
 #import "iTermTouchBarButton.h"
+#import "iTermTouchbarMappings.h"
+#import "iTermVariableScope+Tab.h"
 #import "PTYTab.h"
 
 static NSString *const iTermTabBarTouchBarIdentifier = @"tab bar";
@@ -151,12 +155,76 @@ ITERM_IGNORE_PARTIAL_BEGIN
                           iTermTouchBarIdentifierPreviousMark,
                           iTermTouchBarIdentifierAutocomplete,
                           iTermTouchBarIdentifierStatus ];
-        ids = [ids arrayByAddingObjectsFromArray:[iTermKeyBindingMgr sortedTouchBarKeysInDictionary:[iTermKeyBindingMgr globalTouchBarMap]]];
+        NSArray<iTermTouchbarItem *> *items = [iTermTouchbarMappings sortedTouchbarItemsInDictionary:[iTermTouchbarMappings globalTouchBarMap]];
+        NSArray<NSString *> *customIDs = [items mapWithBlock:^id(iTermTouchbarItem *anObject) {
+            return anObject.identifier;
+        }];
+        customIDs = [self idsByAddingVersions:customIDs];
+        ids = [ids arrayByAddingObjectsFromArray:customIDs];
         self.touchBar.customizationAllowedItemIdentifiers = ids;
         [self updateTouchBarFunctionKeyLabels];
     }
 }
 
+// Returns the version for a global touch bar identifier by checking if its
+// value has changed and incrementing the version if so, otherwise returning
+// the last version. If it's never been seen before then it's assigned version
+// 0.
+//
+// This scheme is used because the touch bar doesn't recognize changes to
+// existing identifiers. If you modify a custom touch bar button then you also
+// have to change its identifier. Prefs has no notion of versioning because
+// that's never been needed before. So the touch bar code imposes its own
+// notion of versions on it.
+//
+// We have to make sure to remove the version number (if present) before using
+// an identifier from the touch bar APIs.
+//
+// The version table is global because custom touch bar items are global.
+//
+// Versioned identifiers will be saved, which breaks downgrading to an older
+// version of iTerm2. Backward compatibility is preserved because the code to
+// remove a version number does nothing if one is missing.
+- (NSNumber *)versionFor:(NSString *)identifier withValue:(NSDictionary *)binding {
+    assert(binding);
+    static NSMutableDictionary<NSString *, NSDictionary *> *table;
+    static dispatch_once_t onceToken;
+    dispatch_once(&onceToken, ^{
+        table = [[NSMutableDictionary alloc] init];
+    });
+    NSString *const bindingKey = @"binding";
+    NSString *const versionKey = @"version";
+    NSDictionary *const entry = table[identifier];
+    NSDictionary *const existingBinding = entry[bindingKey];
+    if (existingBinding && [existingBinding isEqual:binding]) {
+        return entry[versionKey];
+    }
+    if (!existingBinding) {
+        table[identifier] = @{ bindingKey: binding, versionKey: @0 };
+        return @0;
+    }
+    NSNumber *const lastVersion = entry[versionKey];
+    NSNumber *const currentVersion = @(lastVersion.integerValue + 1);
+    table[identifier] = @{ bindingKey: binding, versionKey: currentVersion };
+    return currentVersion;
+}
+
+- (NSArray<NSString *> *)idsByAddingVersions:(NSArray<NSString *> *)unversioned {
+    NSDictionary *map = [iTermTouchbarMappings globalTouchBarMap];
+    return [unversioned mapWithBlock:^id(NSString *string) {
+        NSDictionary *binding = map[string];
+        return [string stringByAppendingFormat:@"/v%@", [self versionFor:string withValue:binding]];
+    }];
+}
+
+- (NSString *)idByRemovingVersion:(NSString *)identifier {
+    NSRange range = [identifier rangeOfString:@"/v"];
+    if (range.location == NSNotFound) {
+        return identifier;
+    }
+    return [identifier substringToIndex:range.location];
+}
+
 - (NSTouchBar *)amendTouchBar:(NSTouchBar *)touchBar {
     if (!IsTouchBarAvailable()) {
         return nil;
@@ -200,7 +268,7 @@ ITERM_IGNORE_PARTIAL_BEGIN
     NSScrollView *scrollView = [[[NSScrollView alloc] init] autorelease];
     NSCustomTouchBarItem *item = [[[NSCustomTouchBarItem alloc] initWithIdentifier:iTermTouchBarFunctionKeysScrollView] autorelease];
     item.view = scrollView;
-    NSView *documentView = [[NSView alloc] init];
+    NSView *documentView = [[[NSView alloc] init] autorelease];
     documentView.translatesAutoresizingMaskIntoConstraints = NO;
     scrollView.documentView = documentView;
     NSButton *previous = nil;
@@ -311,7 +379,7 @@ ITERM_IGNORE_PARTIAL_BEGIN
     NSScrollView *scrollView = [[[NSScrollView alloc] init] autorelease];
     NSCustomTouchBarItem *item = [[[NSCustomTouchBarItem alloc] initWithIdentifier:iTermTouchBarIdentifierColorPresetScrollview] autorelease];
     item.view = scrollView;
-    NSView *documentView = [[NSView alloc] init];
+    NSView *documentView = [[[NSView alloc] init] autorelease];
     documentView.translatesAutoresizingMaskIntoConstraints = NO;
     scrollView.documentView = documentView;
     [self addButtonsToColorPresetsDocumentView:documentView];
@@ -324,7 +392,7 @@ ITERM_IGNORE_PARTIAL_BEGIN
     if ([identifier isEqualToString:iTermTabBarTouchBarIdentifier]) {
         NSScrubber *scrubber;
         if (!self.tabsTouchBarItem) {
-            self.tabsTouchBarItem = [[NSCustomTouchBarItem alloc] initWithIdentifier:identifier];
+            self.tabsTouchBarItem = [[[NSCustomTouchBarItem alloc] initWithIdentifier:identifier] autorelease];
             self.tabsTouchBarItem.customizationLabel = @"Full Screen Tab Bar";
 
             scrubber = [[NSScrubber alloc] initWithFrame:NSMakeRect(0, 0, 320, 30)];
@@ -335,7 +403,7 @@ ITERM_IGNORE_PARTIAL_BEGIN
             [scrubber registerClass:[NSScrubberTextItemView class] forItemIdentifier:iTermTabBarItemTouchBarIdentifier];
 
             // Use the flow layout.
-            NSScrubberLayout *scrubberLayout = [[NSScrubberFlowLayout alloc] init];
+            NSScrubberLayout *scrubberLayout = [[[NSScrubberFlowLayout alloc] init] autorelease];
             scrubber.scrubberLayout = scrubberLayout;
 
             scrubber.mode = NSScrubberModeFree;
@@ -347,7 +415,7 @@ ITERM_IGNORE_PARTIAL_BEGIN
         } else {
             scrubber = (NSScrubber *)self.tabsTouchBarItem.view;
         }
-        // Reload the scrubber after a spin of the runloop bacause it gets laid out tighter after the
+        // Reload the scrubber after a spin of the runloop because it gets laid out tighter after the
         // rest of the toolbar is created. If we reloadData now then it jumps the first time we change
         // tabs.
         dispatch_async(dispatch_get_main_queue(), ^{
@@ -358,7 +426,7 @@ ITERM_IGNORE_PARTIAL_BEGIN
 
         return self.tabsTouchBarItem;
     } else if ([identifier isEqualToString:iTermTouchBarIdentifierAutocomplete]) {
-        self.autocompleteCandidateListItem = [[NSCandidateListTouchBarItem alloc] initWithIdentifier:identifier];
+        self.autocompleteCandidateListItem = [[[NSCandidateListTouchBarItem alloc] initWithIdentifier:identifier] autorelease];
         self.autocompleteCandidateListItem.delegate = self;
         self.autocompleteCandidateListItem.customizationLabel = @"Autocomplete Suggestions";
         NSAttributedString *(^commandUseToAttributedString)(NSString *commandUse,
@@ -381,7 +449,7 @@ ITERM_IGNORE_PARTIAL_BEGIN
         selector = @selector(statusTouchBarItemSelected:);
         label = @"Your Message Here";
     } else if ([identifier isEqualToString:iTermTouchBarIdentifierAddMark]) {
-        image = [[NSImage imageNamed:@"Add Mark Touch Bar Icon"] imageWithColor:[NSColor labelColor]];
+        image = [[NSImage it_imageNamed:@"Add Mark Touch Bar Icon" forClass:self.class] imageWithColor:[NSColor labelColor]];
         selector = @selector(addMarkTouchBarItemSelected:);
         label = @"Add Mark";
     } else if ([identifier isEqualToString:iTermTouchBarIdentifierNextMark]) {
@@ -394,7 +462,6 @@ ITERM_IGNORE_PARTIAL_BEGIN
         label = @"Previous Mark";
     } else if ([identifier isEqualToString:iTermTouchBarIdentifierColorPreset]) {
         image = [NSImage imageNamed:NSImageNameTouchBarColorPickerFill];
-        selector = @selector(colorPresetTouchBarItemSelected:);
         NSPopoverTouchBarItem *item = [[[NSPopoverTouchBarItem alloc] initWithIdentifier:identifier] autorelease];
         item.customizationLabel = @"Color Preset";
         item.showsCloseButton = YES;
@@ -406,7 +473,7 @@ ITERM_IGNORE_PARTIAL_BEGIN
         item.popoverTouchBar = secondaryTouchBar;
         return item;
     } else if ([identifier isEqualToString:iTermTouchBarIdentifierFunctionKeys]) {
-        image = [NSImage imageNamed:@"Touch Bar Function Keys"];
+        image = [NSImage it_imageNamed:@"Touch Bar Function Keys" forClass:self.class];
         NSPopoverTouchBarItem *item = [[[NSPopoverTouchBarItem alloc] initWithIdentifier:identifier] autorelease];
         item.customizationLabel = @"Function Keys Popover";
         item.showsCloseButton = YES;
@@ -435,35 +502,52 @@ ITERM_IGNORE_PARTIAL_BEGIN
         item.customizationLabel = label;
         if ([identifier isEqualToString:iTermTouchBarIdentifierManPage]) {
             button.title = @"";
-            button.image = [NSImage imageNamed:@"Man Page Touch Bar Icon"];
+            button.image = [NSImage it_imageNamed:@"Man Page Touch Bar Icon" forClass:self.class];
             button.imagePosition = NSImageOnly;
             button.enabled = NO;
         }
         return item;
     }
-    NSDictionary *map = [iTermKeyBindingMgr globalTouchBarMap];
-    NSDictionary *binding = map[identifier];
+    
+    // Custom action
+    NSString *unversionedIdentifier = [self idByRemovingVersion:identifier];
+    NSDictionary *map = [iTermTouchbarMappings globalTouchBarMap];
+    NSDictionary *binding = map[unversionedIdentifier];
 
     if (!binding) {
         return nil;
     }
+    iTermKeyBindingAction *action = [iTermKeyBindingAction withDictionary:binding];
+    if (!action) {
+        return nil;
+    }
     NSCustomTouchBarItem *item = [[[NSCustomTouchBarItem alloc] initWithIdentifier:identifier] autorelease];
-    iTermTouchBarButton *button = [iTermTouchBarButton buttonWithTitle:[iTermKeyBindingMgr touchBarLabelForBinding:binding]
+    iTermTouchBarButton *button = [iTermTouchBarButton buttonWithTitle:action.label
                                                                 target:self
                                                                 action:@selector(touchBarItemSelected:)];
     button.keyBindingAction = binding;
     item.view = button;
     item.view.identifier = identifier;
-    item.customizationLabel = [iTermKeyBindingMgr formatAction:binding];
+    item.customizationLabel = action.displayName;
 
     return item;
 }
 
 - (void)touchBarItemSelected:(iTermTouchBarButton *)sender {
-    NSDictionary *binding = sender.keyBindingAction;
-    [self.currentSession performKeyBindingAction:[iTermKeyBindingMgr actionForTouchBarItemBinding:binding]
-                                       parameter:[iTermKeyBindingMgr parameterForTouchBarItemBinding:binding]
-                                           event:[NSApp currentEvent]];
+    NSDictionary *map = [iTermTouchbarMappings globalTouchBarMap];
+    NSString *unversionedIdentifier = [self idByRemovingVersion:sender.identifier];
+    NSDictionary *binding = unversionedIdentifier ? map[unversionedIdentifier] : nil;
+    if (!binding) {
+        // The item has been deleted from prefs but the touch bar retains
+        // existing config until the user edits it.
+        binding = sender.keyBindingAction;
+    }
+    iTermKeyBindingAction *action = [iTermKeyBindingAction withDictionary:binding];
+    if (!action) {
+        DLog(@"Bogus binding: %@", binding);
+        return;
+    }
+    [self.currentSession performKeyBindingAction:action event:[NSApp currentEvent]];
 }
 
 - (void)addMarkTouchBarItemSelected:(id)sender {
@@ -471,17 +555,24 @@ ITERM_IGNORE_PARTIAL_BEGIN
 }
 
 - (void)nextMarkTouchBarItemSelected:(id)sender {
-    [self.currentSession nextMarkOrNote];
+    [self.currentSession nextMark];
 }
 
 - (void)previousMarkTouchBarItemSelected:(id)sender {
-    [self.currentSession previousMarkOrNote];
+    [self.currentSession previousMark];
 }
 
 - (void)manPageTouchBarItemSelected:(iTermTouchBarButton *)sender {
     NSString *command = sender.keyBindingAction[@"command"];
     if (command) {
-        [[iTermController sharedInstance] openSingleUseWindowWithCommand:command];
+        [[iTermController sharedInstance] openSingleUseWindowWithCommand:command
+                                                                  inject:nil
+                                                             environment:nil
+                                                                     pwd:nil
+                                                                 options:(iTermSingleUseWindowOptionsDoNotEscapeArguments |
+                                                                          iTermSingleUseWindowOptionsCloseOnTermination)
+                                                          didMakeSession:nil
+                                                              completion:nil];
     }
 }
 
@@ -532,9 +623,9 @@ ITERM_IGNORE_PARTIAL_BEGIN
     };
     NSString *chars = [NSString stringWithFormat:@"%C", (unichar)(NSF1FunctionKey + number - 1)];
     NSPoint screenPoint = [NSEvent mouseLocation];
-    NSEvent *event = [NSEvent keyEventWithType:NSKeyDown
+    NSEvent *event = [NSEvent keyEventWithType:NSEventTypeKeyDown
                                       location:[self.window convertRectFromScreen:NSMakeRect(screenPoint.x, screenPoint.y, 0, 0)].origin
-                                 modifierFlags:([NSEvent modifierFlags] | NSFunctionKeyMask)
+                                 modifierFlags:([NSEvent modifierFlags] | NSEventModifierFlagFunction)
                                      timestamp:[currentEvent timestamp]
                                   windowNumber:self.window.windowNumber
                                        context:nil
@@ -555,7 +646,7 @@ ITERM_IGNORE_PARTIAL_BEGIN
     if (index != NSNotFound) {
         NSString *command = [anItem candidates][index];
         NSString *prefix = self.currentSession.currentCommand;
-        if ([command hasPrefix:prefix] || prefix.length == 0) {
+        if (prefix.length == 0 || [command hasPrefix:prefix]) {
             [self.currentSession insertText:[command substringFromIndex:prefix.length]];
         }
     }
@@ -575,7 +666,8 @@ ITERM_IGNORE_PARTIAL_BEGIN
 
 - (NSString *)scrubber:(NSScrubber *)scrubber labelAtIndex:(NSInteger)index {
     NSArray<PTYTab *> *tabs = self.tabs;
-    return index < tabs.count ?  self.tabs[index].activeSession.name : @"";
+    PTYTab *tab = index < tabs.count ? self.tabs[index] : nil;
+    return tab.variablesScope.title ?: @"";
 }
 
 - (__kindof NSScrubberItemView *)scrubber:(NSScrubber *)scrubber viewForItemAtIndex:(NSInteger)index {
