@@ -29,23 +29,31 @@
  */
 
 #import "ScreenChar.h"
+
+#import "DebugLogging.h"
 #import "charmaps.h"
 #import "iTermAdvancedSettingsModel.h"
 #import "iTermImageInfo.h"
+#import "iTermMalloc.h"
+#import "NSArray+iTerm.h"
 #import "NSCharacterSet+iTerm.h"
 
 static NSString *const kScreenCharComplexCharMapKey = @"Complex Char Map";
+static NSString *const kScreenCharSpacingCombiningMarksKey = @"Spacing Combining Marks";
 static NSString *const kScreenCharInverseComplexCharMapKey = @"Inverse Complex Char Map";
 static NSString *const kScreenCharImageMapKey = @"Image Map";
 static NSString *const kScreenCharCCMNextKeyKey = @"Next Key";
 static NSString *const kScreenCharHasWrappedKey = @"Has Wrapped";
 
+static NSInteger gScreenCharGeneration;
+
 // Maps codes to strings
 static NSMutableDictionary* complexCharMap;
+static NSMutableSet<NSNumber *> *spacingCombiningMarkCodeNumbers;
 // Maps strings to codes.
 static NSMutableDictionary* inverseComplexCharMap;
 // Image info. Maps a NSNumber with the image's code to an ImageInfo object.
-static NSMutableDictionary* gImages;
+static NSMutableDictionary<NSNumber *, iTermImageInfo *> *gImages;
 static NSMutableDictionary* gEncodableImageMap;
 // Next available code.
 static int ccmNextKey = 1;
@@ -53,6 +61,16 @@ static int ccmNextKey = 1;
 // strings before creating a new one with a recycled code.
 static BOOL hasWrapped = NO;
 
+typedef NS_ENUM(int, iTermTriState) {
+    iTermTriStateFalse,
+    iTermTriStateTrue,
+    iTermTriStateOther
+};
+
+iTermTriState iTermTriStateFromBool(BOOL b) {
+    return b ? iTermTriStateTrue : iTermTriStateFalse;
+}
+
 @interface iTermStringLine()
 @property(nonatomic, retain) NSString *stringValue;
 @end
@@ -125,25 +143,42 @@ static BOOL hasWrapped = NO;
 @synthesize line = _line;
 @synthesize length = _length;
 @synthesize eol = _eol;
+
+- (instancetype)initWithLine:(screen_char_t *)line
+                      length:(int)length
+                continuation:(screen_char_t)continuation {
+    self = [super init];
+    if (self) {
+        _line = line;
+        _length = length;
+        _continuation = continuation;
+        _eol = continuation.code;
+    }
+    return self;
+}
+
+- (BOOL)isEqualToScreenCharArray:(ScreenCharArray *)other {
+    if (!other) {
+        return NO;
+    }
+    return (_line == other->_line &&
+            _length == other->_length &&
+            _eol == other->_eol &&
+            !memcmp(&_continuation, &other->_continuation, sizeof(_continuation)));
+}
+
 @end
 
 static void CreateComplexCharMapIfNeeded() {
     if (!complexCharMap) {
         complexCharMap = [[NSMutableDictionary alloc] initWithCapacity:1000];
-        // Add box-drawing chars, which are reserved. They are drawn using
-        // bezier paths but it's important that the keys refer to an existing
-        // string for general correctness.
-        for (int i = 0; i < 256; i++) {
-            if (lineDrawingCharFlags[i]) {
-                complexCharMap[@(charmap[i])] = [NSString stringWithFormat:@"%C", charmap[i]];
-            }
-        }
+        spacingCombiningMarkCodeNumbers = [[NSMutableSet alloc] initWithCapacity:1000];
         inverseComplexCharMap = [[NSMutableDictionary alloc] initWithCapacity:1000];
+        gScreenCharGeneration++;
     }
 }
 
-NSString* ComplexCharToStr(int key)
-{
+NSString *ComplexCharToStr(int key) {
     if (key == UNICODE_REPLACEMENT_CHAR) {
         return ReplacementString();
     }
@@ -152,13 +187,15 @@ NSString* ComplexCharToStr(int key)
     return [complexCharMap objectForKey:[NSNumber numberWithInt:key]];
 }
 
-NSString* ScreenCharToStr(screen_char_t* sct)
-{
+BOOL ComplexCharCodeIsSpacingCombiningMark(unichar code) {
+    return [spacingCombiningMarkCodeNumbers containsObject:@(code)];
+}
+
+NSString *ScreenCharToStr(const screen_char_t *const sct) {
     return CharToStr(sct->code, sct->complexChar);
 }
 
-NSString* CharToStr(unichar code, BOOL isComplex)
-{
+NSString *CharToStr(unichar code, BOOL isComplex) {
     if (code == UNICODE_REPLACEMENT_CHAR) {
         return ReplacementString();
     }
@@ -201,29 +238,14 @@ UTF32Char CharToLongChar(unichar code, BOOL isComplex)
 }
 
 static BOOL ComplexCharKeyIsReserved(int k) {
-    switch (k) {
-        case ITERM_BOX_DRAWINGS_LIGHT_UP_AND_LEFT:
-        case ITERM_BOX_DRAWINGS_LIGHT_DOWN_AND_LEFT:
-        case ITERM_BOX_DRAWINGS_LIGHT_DOWN_AND_RIGHT:
-        case ITERM_BOX_DRAWINGS_LIGHT_UP_AND_RIGHT:
-        case ITERM_BOX_DRAWINGS_LIGHT_VERTICAL_AND_HORIZONTAL:
-        case ITERM_BOX_DRAWINGS_LIGHT_HORIZONTAL:
-        case ITERM_BOX_DRAWINGS_LIGHT_VERTICAL_AND_RIGHT:
-        case ITERM_BOX_DRAWINGS_LIGHT_VERTICAL_AND_LEFT:
-        case ITERM_BOX_DRAWINGS_LIGHT_UP_AND_HORIZONTAL:
-        case ITERM_BOX_DRAWINGS_LIGHT_DOWN_AND_HORIZONTAL:
-        case ITERM_BOX_DRAWINGS_LIGHT_VERTICAL:
-            return YES;
-
-        default:
-            return NO;
-    }
+    return k >= iTermBoxDrawingCodeMin && k <= iTermBoxDrawingCodeMax;
 }
 
 static void AllocateImageMapsIfNeeded(void) {
     if (!gImages) {
         gImages = [[NSMutableDictionary alloc] init];
         gEncodableImageMap = [[NSMutableDictionary alloc] init];
+        gScreenCharGeneration++;
     }
 }
 
@@ -238,7 +260,8 @@ screen_char_t ImageCharForNewImage(NSString *name,
         newKey = ccmNextKey++;
     } while (ComplexCharKeyIsReserved(newKey));
 
-    screen_char_t c = { 0 };
+    screen_char_t c;
+    memset(&c, 0, sizeof(c));
     c.image = 1;
     c.code = newKey;
 
@@ -248,6 +271,8 @@ screen_char_t ImageCharForNewImage(NSString *name,
     imageInfo.size = NSMakeSize(width, height);
     imageInfo.inset = inset;
     gImages[@(c.code)] = imageInfo;
+    gScreenCharGeneration++;
+    DLog(@"Assign %@ to image code %@", imageInfo, @(c.code));
 
     return c;
 }
@@ -258,15 +283,22 @@ void SetPositionInImageChar(screen_char_t *charPtr, int x, int y)
     charPtr->backgroundColor = y;
 }
 
-void SetDecodedImage(unichar code, NSImage *image, NSData *data) {
+void SetDecodedImage(unichar code, iTermImage *image, NSData *data) {
     iTermImageInfo *imageInfo = gImages[@(code)];
     [imageInfo setImageFromImage:image data:data];
     gEncodableImageMap[@(code)] = [imageInfo dictionary];
+    gScreenCharGeneration++;
+    if ([iTermAdvancedSettingsModel restoreWindowContents]) {
+        [NSApp invalidateRestorableState];
+    }
+    DLog(@"set decoded image in %@", imageInfo);
 }
 
 void ReleaseImage(unichar code) {
+    DLog(@"ReleaseImage(%@)", @(code));
     [gImages removeObjectForKey:@(code)];
     [gEncodableImageMap removeObjectForKey:@(code)];
+    gScreenCharGeneration++;
 }
 
 iTermImageInfo *GetImageInfo(unichar code) {
@@ -278,14 +310,15 @@ VT100GridCoord GetPositionOfImageInChar(screen_char_t c) {
                               c.backgroundColor);
 }
 
-int GetOrSetComplexChar(NSString* str)
-{
+int GetOrSetComplexChar(NSString *str,
+                        iTermTriState isSpacingCombiningMark) {
     CreateComplexCharMapIfNeeded();
-    NSNumber* number = [inverseComplexCharMap objectForKey:str];
+    NSNumber *number = inverseComplexCharMap[str];
     if (number) {
         return [number intValue];
     }
 
+    gScreenCharGeneration++;
     int newKey;
     do {
         newKey = ccmNextKey++;
@@ -296,6 +329,20 @@ int GetOrSetComplexChar(NSString* str)
         NSString* oldStr = complexCharMap[number];
         if (oldStr) {
             [inverseComplexCharMap removeObjectForKey:oldStr];
+            [spacingCombiningMarkCodeNumbers removeObject:number];
+        }
+    }
+    switch (isSpacingCombiningMark) {
+        case iTermTriStateTrue:
+            [spacingCombiningMarkCodeNumbers addObject:number];
+            break;
+        case iTermTriStateFalse:
+            break;
+        case iTermTriStateOther: {
+            NSCharacterSet *scmSet = [NSCharacterSet spacingCombiningMarksForUnicodeVersion:12];
+            if ([str rangeOfCharacterFromSet:scmSet].location != NSNotFound) {
+                [spacingCombiningMarkCodeNumbers addObject:number];
+            }
         }
     }
     complexCharMap[number] = str;
@@ -310,8 +357,7 @@ int GetOrSetComplexChar(NSString* str)
     return newKey;
 }
 
-int AppendToComplexChar(int key, unichar codePoint)
-{
+int AppendToComplexChar(int key, unichar codePoint) {
     if (key == UNICODE_REPLACEMENT_CHAR) {
         return UNICODE_REPLACEMENT_CHAR;
     }
@@ -326,33 +372,38 @@ int AppendToComplexChar(int key, unichar codePoint)
     NSMutableString* temp = [NSMutableString stringWithString:str];
     [temp appendString:[NSString stringWithCharacters:&codePoint length:1]];
 
-    return GetOrSetComplexChar(temp);
+    return GetOrSetComplexChar(temp, iTermTriStateOther);
 }
 
-void BeginComplexChar(screen_char_t *screenChar, unichar combiningChar, BOOL useHFSPlusMapping) {
-    unichar initialCodePoint = screenChar->code;
-    if (initialCodePoint == UNICODE_REPLACEMENT_CHAR) {
-        return;
+NSString *StringByNormalizingString(NSString *theString, iTermUnicodeNormalization normalization) {
+    NSString *normalizedString;
+    switch (normalization) {
+        case iTermUnicodeNormalizationNFC:
+            normalizedString = [theString precomposedStringWithCanonicalMapping];
+            break;
+        case iTermUnicodeNormalizationNFD:
+            normalizedString = [theString decomposedStringWithCanonicalMapping];
+            break;
+        case iTermUnicodeNormalizationNone:
+            normalizedString = theString;
+            break;
+        case iTermUnicodeNormalizationHFSPlus:
+            normalizedString = [theString precomposedStringWithHFSPlusMapping];
+            break;
     }
-
-    unichar temp[2];
-    temp[0] = initialCodePoint;
-    temp[1] = combiningChar;
-
-    // See if it makes a single code in NFC.
-    NSString *theString = [NSString stringWithCharacters:temp length:2];
-    SetComplexCharInScreenChar(screenChar, theString, useHFSPlusMapping);
+    return normalizedString;
 }
 
 void SetComplexCharInScreenChar(screen_char_t *screenChar,
                                 NSString *theString,
-                                BOOL useHFSPlusMapping) {
-    NSString *nfc = useHFSPlusMapping ? [theString precomposedStringWithHFSPlusMapping] :
-                                        [theString precomposedStringWithCanonicalMapping];
-    if (nfc.length == 1) {
-        screenChar->code = [nfc characterAtIndex:0];
+                                iTermUnicodeNormalization normalization,
+                                BOOL isSpacingCombiningMark) {
+    NSString *normalizedString = StringByNormalizingString(theString, normalization);
+    [theString precomposedStringWithCanonicalMapping];
+    if (normalizedString.length == 1 && !isSpacingCombiningMark) {
+        screenChar->code = [normalizedString characterAtIndex:0];
     } else {
-        screenChar->code = GetOrSetComplexChar(theString);
+        screenChar->code = GetOrSetComplexChar(theString, iTermTriStateFromBool(isSpacingCombiningMark));
         screenChar->complexChar = YES;
     }
 }
@@ -410,9 +461,9 @@ NSString* ScreenCharArrayToString(screen_char_t* screenChars,
                                   unichar** backingStorePtr,
                                   int** deltasPtr) {
     const int lineLength = end - start;
-    unichar* charHaystack = malloc(sizeof(unichar) * lineLength * kMaxParts + 1);
+    unichar* charHaystack = iTermMalloc(sizeof(unichar) * lineLength * kMaxParts + 1);
     *backingStorePtr = charHaystack;
-    int* deltas = malloc(sizeof(int) * (lineLength * kMaxParts + 1));
+    int* deltas = iTermMalloc(sizeof(int) * (lineLength * kMaxParts + 1));
     *deltasPtr = deltas;
     // The 'deltas' array gives the difference in position between the screenChars
     // and the charHaystack. The formula to convert an index in the charHaystack
@@ -450,7 +501,9 @@ NSString* ScreenCharArrayToString(screen_char_t* screenChars,
     int o = 0;
     for (int i = start; i < end; ++i) {
         unichar c = screenChars[i].code;
-        if (c == DWC_RIGHT) {
+        if (c >= ITERM2_PRIVATE_BEGIN && c <= ITERM2_PRIVATE_END) {
+            // Skip private-use characters which signify things like double-width characters and
+            // tab fillers.
             ++delta;
         } else {
             const int len = ExpandScreenChar(&screenChars[i], charHaystack + o);
@@ -498,7 +551,7 @@ NSString* ScreenCharArrayToStringDebug(screen_char_t* screenChars,
     for (int i = 0; i < lineLength; ++i) {
         unichar c = screenChars[i].code;
         if (c != 0 && c != DWC_RIGHT) {
-            [result appendString:ScreenCharToStr(&screenChars[i])];
+            [result appendString:ScreenCharToStr(&screenChars[i]) ?: @"😮"];
         }
     }
     return result;
@@ -513,6 +566,16 @@ int EffectiveLineLength(screen_char_t* theLine, int totalLength) {
     return 0;
 }
 
+NSString *DebugStringForScreenChar(screen_char_t c) {
+    NSArray *modes = @[ @"default", @"selected", @"altsem", @"altsem-reversed" ];
+    return [NSString stringWithFormat:@"<screen_char_t: code=%@ complex=%@ image=%@ url=%@ foregroundColor=%@ fgGreen=%@ fgBlue=%@ backgroundColor=%@ bgGreen=%@ bgBlue=%@ fgMode=%@ bgMode=%@ bold=%@ faint=%@ italic=%@ blink=%@ underline=%@ strikethrough=%@ underlinestyle=%@ unused=%@>",
+            @(c.code), @(c.complexChar), @(c.image), @(c.urlCode),
+            @(c.foregroundColor), @(c.fgGreen), @(c.fgBlue),
+            @(c.backgroundColor), @(c.bgGreen), @(c.bgBlue), modes[c.foregroundColorMode],
+            modes[c.backgroundColorMode], @(c.bold), @(c.faint), @(c.italic), @(c.blink),
+            @(c.underline), @(c.strikethrough), @(c.underlineStyle), @(c.unused)];
+}
+
 // Convert a string into an array of screen characters, dealing with surrogate
 // pairs, combining marks, nonspacing marks, and double-width characters.
 void StringToScreenChars(NSString *s,
@@ -523,10 +586,13 @@ void StringToScreenChars(NSString *s,
                          BOOL ambiguousIsDoubleWidth,
                          int* cursorIndex,
                          BOOL *foundDwc,
-                         BOOL useHFSPlusMapping) {
+                         iTermUnicodeNormalization normalization,
+                         NSInteger unicodeVersion) {
     __block NSInteger j = 0;
     __block BOOL foundCursor = NO;
-    NSCharacterSet *zeroWidthSpaces = [NSCharacterSet zeroWidthSpaceCharacterSet];
+    NSCharacterSet *zeroWidthSpaces = [NSCharacterSet zeroWidthSpaceCharacterSetForUnicodeVersion:unicodeVersion];
+    NSCharacterSet *spacingCombiningMarks = [NSCharacterSet spacingCombiningMarksForUnicodeVersion:12];
+
     [s enumerateComposedCharacters:^(NSRange range,
                                      unichar baseBmpChar,
                                      NSString *composedOrNonBmpChar,
@@ -537,7 +603,7 @@ void StringToScreenChars(NSString *s,
         }
 
         BOOL isDoubleWidth = NO;
-
+        BOOL spacingCombiningMark = NO;
         InitializeScreenChar(buf + j, fg, bg);
 
         // Set the code and the complex flag. Also return early if no cell should be used by this
@@ -546,6 +612,10 @@ void StringToScreenChars(NSString *s,
             if ([zeroWidthSpaces characterIsMember:baseBmpChar]) {
                 // Ignore zero-width spacers.
                 return;
+            } else if ([spacingCombiningMarks characterIsMember:baseBmpChar]) {
+                composedOrNonBmpChar = [NSString stringWithLongCharacter:baseBmpChar];
+                baseBmpChar = 0;
+                spacingCombiningMark = YES;
             } else if (baseBmpChar >= ITERM2_PRIVATE_BEGIN && baseBmpChar <= ITERM2_PRIVATE_END) {
                 // Convert private range characters into the replacement character.
                 baseBmpChar = UNICODE_REPLACEMENT_CHAR;
@@ -556,19 +626,46 @@ void StringToScreenChars(NSString *s,
                 // High surrogate not followed by low surrogate.
                 baseBmpChar = UNICODE_REPLACEMENT_CHAR;
             }
-            buf[j].code = baseBmpChar;
-            buf[j].complexChar = NO;
+            if (!composedOrNonBmpChar) {
+                buf[j].code = baseBmpChar;
+                buf[j].complexChar = NO;
 
-            isDoubleWidth = [NSString isDoubleWidthCharacter:baseBmpChar
-                                      ambiguousIsDoubleWidth:ambiguousIsDoubleWidth];
-        } else {
-            SetComplexCharInScreenChar(buf + j, composedOrNonBmpChar, useHFSPlusMapping);
+                isDoubleWidth = [NSString isDoubleWidthCharacter:baseBmpChar
+                                          ambiguousIsDoubleWidth:ambiguousIsDoubleWidth
+                                                  unicodeVersion:unicodeVersion];
+            }
+        }
+        if (composedOrNonBmpChar) {
+            const NSUInteger composedLength = composedOrNonBmpChar.length;
+            // Ensure the string is not longer than what we support.
+            if (composedLength > kMaxParts) {
+                composedOrNonBmpChar = [composedOrNonBmpChar substringToIndex:kMaxParts];
+
+                // Ensure a high surrogate isn't left dangling at the end.
+                if (CFStringIsSurrogateHighCharacter([composedOrNonBmpChar characterAtIndex:kMaxParts - 1])) {
+                    composedOrNonBmpChar = [composedOrNonBmpChar substringToIndex:kMaxParts - 1];
+                }
+            }
+            SetComplexCharInScreenChar(buf + j, composedOrNonBmpChar, normalization, spacingCombiningMark);
+            NSInteger next = 1;
             UTF32Char baseChar = [composedOrNonBmpChar characterAtIndex:0];
-            if (IsHighSurrogate(baseChar) && composedOrNonBmpChar.length > 1) {
+            if (IsHighSurrogate(baseChar) && composedLength > 1) {
                 baseChar = DecodeSurrogatePair(baseChar, [composedOrNonBmpChar characterAtIndex:1]);
+                next += 1;
             }
             isDoubleWidth = [NSString isDoubleWidthCharacter:baseChar
-                                      ambiguousIsDoubleWidth:ambiguousIsDoubleWidth];
+                                      ambiguousIsDoubleWidth:ambiguousIsDoubleWidth
+                                              unicodeVersion:unicodeVersion];
+            if (!isDoubleWidth && composedLength > next) {
+                const unichar peek = [composedOrNonBmpChar characterAtIndex:next];
+                if (peek == 0xfe0f) {
+                    // VS16
+                    if ([[NSCharacterSet emojiAcceptingVS16] characterIsMember:baseChar] &&
+                        [iTermAdvancedSettingsModel vs16Supported]) {
+                        isDoubleWidth = YES;
+                    }
+                }
+            }
         }
 
         // Append a DWC_RIGHT if the base character is double-width.
@@ -612,8 +709,10 @@ void InitializeScreenChar(screen_char_t *s, screen_char_t fg, screen_char_t bg)
     s->italic = fg.italic;
     s->blink = fg.blink;
     s->underline = fg.underline;
+    s->strikethrough = fg.strikethrough;
+    s->underlineStyle = fg.underlineStyle;
     s->image = NO;
-
+    s->urlCode = fg.urlCode;
     s->unused = 0;
 }
 
@@ -623,34 +722,51 @@ void ConvertCharsToGraphicsCharset(screen_char_t *s, int len)
 
     for (i = 0; i < len; i++) {
         assert(!s[i].complexChar);
-        s[i].complexChar = lineDrawingCharFlags[(int)(s[i].code)];
+        s[i].complexChar = NO;
         s[i].code = charmap[(int)(s[i].code)];
     }
 }
 
+NSInteger ScreenCharGeneration(void) {
+    return gScreenCharGeneration;
+}
+
 NSDictionary *ScreenCharEncodedRestorableState(void) {
     return @{ kScreenCharComplexCharMapKey: complexCharMap ?: @{},
+              kScreenCharSpacingCombiningMarksKey: spacingCombiningMarkCodeNumbers.allObjects ?: @[],
               kScreenCharInverseComplexCharMapKey: inverseComplexCharMap ?: @{},
               kScreenCharImageMapKey: gEncodableImageMap ?: @{},
               kScreenCharCCMNextKeyKey: @(ccmNextKey),
               kScreenCharHasWrappedKey: @(hasWrapped) };
 }
 
+void ScreenCharGarbageCollectImages(void) {
+    NSArray<NSNumber *> *provisionalKeys = [gImages.allKeys filteredArrayUsingBlock:^BOOL(NSNumber *key) {
+        return gImages[key].provisional;
+    }];
+    DLog(@"Garbage collect: %@", provisionalKeys);
+    [gImages removeObjectsForKeys:provisionalKeys];
+}
+
+void ScreenCharClearProvisionalFlagForImageWithCode(int code) {
+    DLog(@"Clear provisional for %@", @(code));
+    gImages[@(code)].provisional = NO;
+}
+
 void ScreenCharDecodeRestorableState(NSDictionary *state) {
     NSDictionary *stateComplexCharMap = state[kScreenCharComplexCharMapKey];
-    if (!complexCharMap && stateComplexCharMap.count) {
-        complexCharMap = [[NSMutableDictionary alloc] init];
-    }
+    CreateComplexCharMapIfNeeded();
     for (id key in stateComplexCharMap) {
         if (!complexCharMap[key]) {
             complexCharMap[key] = stateComplexCharMap[key];
         }
     }
+    NSArray<NSString *> *spacingCombiningMarksArray = state[kScreenCharSpacingCombiningMarksKey];
+    for (NSNumber *number in spacingCombiningMarksArray) {
+        [spacingCombiningMarkCodeNumbers addObject:number];
+    }
 
     NSDictionary *stateInverseMap = state[kScreenCharInverseComplexCharMapKey];
-    if (!inverseComplexCharMap && stateInverseMap.count) {
-        inverseComplexCharMap = [[NSMutableDictionary alloc] init];
-    }
     for (id key in stateInverseMap) {
         if (!inverseComplexCharMap[key]) {
             inverseComplexCharMap[key] = stateInverseMap[key];
@@ -662,9 +778,107 @@ void ScreenCharDecodeRestorableState(NSDictionary *state) {
         gEncodableImageMap[key] = imageMap[key];
         iTermImageInfo *info = [[[iTermImageInfo alloc] initWithDictionary:imageMap[key]] autorelease];
         if (info) {
+            info.provisional = YES;
             gImages[key] = info;
+            DLog(@"Decoded restorable state for image %@: %@", key, info);
         }
     }
     ccmNextKey = [state[kScreenCharCCMNextKeyKey] intValue];
     hasWrapped = [state[kScreenCharHasWrappedKey] boolValue];
 }
+
+static NSString *ScreenCharColorDescription(unsigned int red,
+                                            unsigned int green,
+                                            unsigned int blue,
+                                            ColorMode mode) {
+    switch (mode) {
+        case ColorModeAlternate:
+            switch (red) {
+                case ALTSEM_DEFAULT:
+                    return @"Default";
+                case ALTSEM_SELECTED:
+                    return @"Selected";
+                case ALTSEM_CURSOR:
+                    return @"Cursor";
+                case ALTSEM_REVERSED_DEFAULT:
+                    return @"Reversed";
+                case ALTSEM_SYSTEM_MESSAGE:
+                    return @"System";
+            }
+            return @"Invalid";
+
+        case ColorModeNormal:
+            if (red < 16) {
+                NSString *color = @"";
+                NSArray<NSString *> *names = @[ @"Black", @"Red", @"Green", @"Yellow", @"Blue",
+                                                @"Magenta", @"Cyan", @"White" ];
+                if (red > 7) {
+                    color = [color stringByAppendingString:@"ANSI Bright "];
+                }
+                color = [color stringByAppendingString:names[red & 7]];
+                return color;
+            }
+            if (red < 232) {
+                const int r = (red - 16) / 36;
+                const int g = ((red - 16) - r*36) / 6;
+                const int b = ((red - 16) - r*36 - g*6);
+                return [NSString stringWithFormat:@"8bit(%@/5,%@/5,%@/5)", @(r), @(g), @(b)];
+            }
+            const int gray = red-232;
+            return [NSString stringWithFormat:@"gray%@/22", @(gray)];
+        case ColorMode24bit:
+            return [NSString stringWithFormat:@"24bit(%@,%@,%@)", @(red), @(green), @(blue)];
+        case ColorModeInvalid:
+            return @"Invalid";
+    }
+    return @"Invalid";
+}
+
+NSString *ScreenCharDescription(screen_char_t c) {
+    if (c.image) {
+        return nil;
+    }
+    NSMutableArray<NSString *> *attrs = [NSMutableArray array];
+    if (c.bold) {
+        [attrs addObject:@"Bold"];
+    }
+    if (c.faint) {
+        [attrs addObject:@"Faint"];
+    }
+    if (c.italic) {
+        [attrs addObject:@"Italic"];
+    }
+    if (c.blink) {
+        [attrs addObject:@"Blink"];
+    }
+    if (c.underline) {
+        switch (c.underlineStyle) {
+            case VT100UnderlineStyleSingle:
+                [attrs addObject:@"Underline"];
+                break;
+            case VT100UnderlineStyleCurly:
+                [attrs addObject:@"Curly-Underline"];
+                break;
+        }
+    }
+    if (c.strikethrough) {
+        [attrs addObject:@"Strike"];
+    }
+    if (c.urlCode) {
+        [attrs addObject:@"URL"];
+    }
+    NSString *style = [attrs componentsJoinedByString:@" "];
+    if (style.length) {
+        style = [@" " stringByAppendingString:style];
+    }
+    return [NSString stringWithFormat:@"fg=%@ bg=%@%@",
+            ScreenCharColorDescription(c.foregroundColor,
+                                       c.fgGreen,
+                                       c.fgBlue,
+                                       c.foregroundColorMode),
+            ScreenCharColorDescription(c.backgroundColor,
+                                       c.bgGreen,
+                                       c.bgBlue,
+                                       c.backgroundColorMode),
+            style];
+}
