@@ -59,7 +59,12 @@ typedef enum {
 
 // Read either an integer followed by a semicolon or letter "P".
 + (iTermXtermParserState)parseModeFromContext:(iTermParserContext *)context mode:(int *)mode {
-    if (iTermParserConsumeInteger(context, mode)) {
+    BOOL overflow = NO;
+    iTermParserContext saved = *context;
+    if (iTermParserConsumeInteger(context, mode, &overflow)) {
+        if (overflow) {
+            return kXtermParserFailingState;
+        }
         // Read an integer. Either out of data or a semicolon should follow; anything else is
         // a malformed input.
         if (iTermParserCanAdvance(context)) {
@@ -72,7 +77,9 @@ typedef enum {
                 return kXtermParserFailingState;
             }
         } else {
-            // Out of data.
+            // Out of data. Backtrack to the start of the int because when this is resumed later
+            // with more data, we must begin parsing the mode all over again.
+            *context = saved;
             return kXtermParserOutOfDataState;
         }
     } else {
@@ -120,8 +127,14 @@ typedef enum {
 //   ESC <anything else>  Fails
 //   File= KVP code       Finish prior to true end of OSC
 //   CAN or SUB           Fails
+//
+// When 8-bit control sequences are supported, the following behaviors are added:
+//   C1 ST                Finished parsing
+//   C1 OSC               This is ignored
+//   Other C1 codes       Fails
 // Other characters are appended to |data|.
 + (iTermXtermParserState)parseNextCharsInStringFromContext:(iTermParserContext *)context
+                              support8BitControlCharacters:(BOOL)support8BitControlCharacters
                                                       data:(NSMutableData *)data
                                                       mode:(int)mode {
     iTermXtermParserState nextState = kXtermParserParsingStringState;
@@ -172,6 +185,39 @@ typedef enum {
                     nextState = kXtermParserFinishedState;
                     break;
 
+                case VT100CC_C1_ST:
+                    if (support8BitControlCharacters) {
+                        nextState = kXtermParserFinishedState;
+                        break;
+                    }
+                    // fall through
+                case VT100CC_C1_OSC:
+                    if (support8BitControlCharacters) {
+                        append = NO;
+                        nextState = kXtermParserParsingStringState;
+                        break;
+                    }
+                    // fall through
+                case VT100CC_C1_IND:
+                case VT100CC_C1_NEL:
+                case VT100CC_C1_HTS:
+                case VT100CC_C1_RI:
+                case VT100CC_C1_SS2:
+                case VT100CC_C1_SS3:
+                case VT100CC_C1_DCS:
+                case VT100CC_C1_SPA:
+                case VT100CC_C1_EPA:
+                case VT100CC_C1_SOS:
+                case VT100CC_C1_DECID:
+                case VT100CC_C1_CSI:
+                case VT100CC_C1_PM:
+                case VT100CC_C1_APC:
+                    if (support8BitControlCharacters) {
+                        nextState = kXtermParserFailedState;
+                        break;
+                    }
+                    // fall through
+
                 default:
                     nextState = kXtermParserParsingStringState;
                     break;
@@ -200,7 +246,7 @@ typedef enum {
                @6: @(XTERMCC_PROPRIETARY_ETERM_EXT),
                @7: @(XTERMCC_PWD_URL),
                @8: @(XTERMCC_LINK),
-               @9: @(ITERM_GROWL),
+               @9: @(ITERM_USER_NOTIFICATION),
                @10: @(XTERMCC_TEXT_FOREGROUND_COLOR),
                @11: @(XTERMCC_TEXT_BACKGROUND_COLOR),
                // 50 is a nonstandard escape code implemented by Konsole.
@@ -231,6 +277,7 @@ typedef enum {
     headerToken.string = [[[NSString alloc] initWithData:data
                                                 encoding:encoding] autorelease];
     [self parseKeyValuePairInToken:headerToken];
+    [headerToken retain];
     CVectorAppend(vector, headerToken);
 }
 
@@ -241,6 +288,7 @@ typedef enum {
     token->type = XTERMCC_MULTITOKEN_BODY;
     token.string = [[[NSString alloc] initWithData:data
                                           encoding:encoding] autorelease];
+    [token retain];
     CVectorAppend(vector, token);
 }
 
@@ -254,6 +302,7 @@ typedef enum {
     int mode = 0;
     iTermXtermParserState state = kXtermParserParsingModeState;
     BOOL multitokenHeaderEmitted = NO;
+    BOOL support8BitControlCharacters = (encoding == NSASCIIStringEncoding || encoding == NSISOLatin1StringEncoding);
 
     if (savedState.count) {
         data = savedState[kXtermParserSavedStateDataKey];
@@ -263,9 +312,22 @@ typedef enum {
         state = [savedState[kXtermParserSavedStateStateKey] intValue];
         multitokenHeaderEmitted = [savedState[kXtermParserMultitokenHeaderEmittedkey] boolValue];
     } else {
-        iTermParserConsumeOrDie(context, VT100CC_ESC);
-        // 99% of the time the next byte is a ], but it could be a _ which is APC (used by tmux).
-        if (iTermParserConsume(context) == '_') {
+        unsigned char peek = iTermParserPeek(context);
+        BOOL apc = NO;
+        if (support8BitControlCharacters && (peek == VT100CC_C1_OSC ||
+                                             peek == VT100CC_C1_APC)) {
+            if (peek == VT100CC_C1_APC) {
+                apc = YES;
+            }
+            iTermParserConsume(context);
+        } else {
+            iTermParserConsumeOrDie(context, VT100CC_ESC);
+            // 99% of the time the next byte is a ], but it could be a _ which is APC (used by tmux).
+            if (iTermParserConsume(context) == '_') {
+                apc = YES;
+            }
+        }
+        if (apc) {
             mode = kAPCMode;
             state = kXtermParserParsingStringState;
         }
@@ -286,7 +348,9 @@ typedef enum {
                 break;
 
             case kXtermParserParsingStringState:
-                state = [self parseNextCharsInStringFromContext:context data:data mode:mode];
+                state = [self parseNextCharsInStringFromContext:context
+                                   support8BitControlCharacters:support8BitControlCharacters
+                                                           data:data mode:mode];
                 break;
 
             case kXtermParserHeaderEndState:
@@ -302,7 +366,10 @@ typedef enum {
                 break;
 
             case kXtermParserFailingState:
-                state = [self parseNextCharsInStringFromContext:context data:nil mode:0];
+                state = [self parseNextCharsInStringFromContext:context
+                                   support8BitControlCharacters:support8BitControlCharacters
+                                                           data:nil
+                                                           mode:0];
 
                 // Convert success states into failure states.
                 if (state == kXtermParserFinishedState) {
@@ -336,7 +403,7 @@ typedef enum {
                     }
                 }
                 return;
-                
+
             case kXtermParserOutOfDataState:
                 if (data.length && multitokenHeaderEmitted) {
                     [self emitIncidentalForMultitokenBodyInVector:incidentals
