@@ -216,16 +216,13 @@
 - (BOOL)fitWindowToTabSize:(NSSize)tabSize;
 
 // Return the index of a tab or NSNotFound.
-// This method is used, for example, in iTermExpose, where PTYTabs are shown
-// side by side, and one needs to determine which index it has, so it can be
-// selected when leaving iTerm expose.
 - (NSInteger)indexOfTab:(PTYTab*)aTab;
 
 // Insert a tab at a specified location.
 - (void)insertTab:(PTYTab*)aTab atIndex:(int)anIndex;
 
 // Add a session to the tab view.
-- (void)insertSession:(PTYSession *)aSession atIndex:(int)anIndex;
+- (PTYTab *)insertSession:(PTYSession *)aSession atIndex:(int)anIndex;
 
 // Resize window to be just large enough to fit the largest tab without
 // changing session sizes.
