@@ -12,11 +12,15 @@
 // Optionally, it may have a function that computes its value (set in +computedObjectDictionary)
 // and the view controller may customize how its control's appearance changes dynamically.
 
+#import "iTermNotificationCenter.h"
+#import "iTermPreferenceDidChangeNotification.h"
 #import "iTermPreferences.h"
 #import "iTermRemotePreferences.h"
+#import "iTermUserDefaultsObserver.h"
 #import "WindowArrangements.h"
+#import "PSMTabBarControl.h"
 
-#define BLOCK(x) [[^id() { return [self x]; } copy] autorelease]
+#define BLOCK(x) [^id() { return [self x]; } copy]
 
 NSString *const kPreferenceKeyOpenBookmark = @"OpenBookmark";
 NSString *const kPreferenceKeyOpenArrangementAtStartup = @"OpenArrangementAtStartup";
@@ -24,6 +28,7 @@ NSString *const kPreferenceKeyOpenNoWindowsAtStartup = @"OpenNoWindowsAtStartup"
 NSString *const kPreferenceKeyQuitWhenAllWindowsClosed = @"QuitWhenAllWindowsClosed";
 NSString *const kPreferenceKeyConfirmClosingMultipleTabs = @"OnlyWhenMoreTabs";  // The key predates split panes
 NSString *const kPreferenceKeyPromptOnQuit = @"PromptOnQuit";
+NSString *const kPreferenceKeyPromptOnQuitEvenIfThereAreNoWindows = @"PromptOnQuitEvenIfThereAreNoWindows";
 NSString *const kPreferenceKeyInstantReplayMemoryMegabytes = @"IRMemory";
 NSString *const kPreferenceKeySavePasteAndCommandHistory = @"SavePasteHistory";  // The key predates command history
 NSString *const kPreferenceKeyAddBonjourHostsToProfiles = @"EnableRendezvous";  // The key predates the name Bonjour
@@ -40,6 +45,8 @@ NSString *const kPreferenceKeyNeverRemindPrefsChangesLostForFileSelection = @"No
 // YES = apply preference from above key, NO = ask on exit if changes exist
 NSString *const kPreferenceKeyNeverRemindPrefsChangesLostForFileHaveSelection = @"NoSyncNeverRemindPrefsChangesLostForFile";
 
+NSString *const iTermMetalSettingsDidChangeNotification = @"iTermMetalSettingsDidChangeNotification";
+
 NSString *const kPreferenceKeyCustomFolder = @"PrefsCustomFolder";
 NSString *const kPreferenceKeySelectionCopiesText = @"CopySelection";
 NSString *const kPreferenceKeyCopyLastNewline = @"CopyLastNewline";
@@ -48,31 +55,55 @@ NSString *const kPreferenceKeyCharactersConsideredPartOfAWordForSelection = @"Wo
 NSString *const kPreferenceKeySmartWindowPlacement = @"SmartPlacement";
 NSString *const kPreferenceKeyAdjustWindowForFontSizeChange = @"AdjustWindowForFontSizeChange";
 NSString *const kPreferenceKeyMaximizeVerticallyOnly = @"MaxVertically";
-NSString *const kPreferenceKeyLionStyleFullscren = @"UseLionStyleFullscreen";
+NSString *const kPreferenceKeyLionStyleFullscreen = @"UseLionStyleFullscreen";
+NSString *const kPreferenceKeySeparateWindowTitlePerTab = @"SeparateWindowTitlePerTab";
 NSString *const kPreferenceKeyOpenTmuxWindowsIn = @"OpenTmuxWindowsIn";
 NSString *const kPreferenceKeyTmuxDashboardLimit = @"TmuxDashboardLimit";
 NSString *const kPreferenceKeyAutoHideTmuxClientSession = @"AutoHideTmuxClientSession";
+NSString *const kPreferenceKeyUseTmuxProfile = @"TmuxUsesDedicatedProfile";
+NSString *const kPreferenceKeyUseTmuxStatusBar = @"UseTmuxStatusBar";
+NSString *const kPreferenceKeyTmuxPauseModeAgeLimit = @"TmuxPauseModeAgeLimit";
+NSString *const kPreferenceKeyTmuxUnpauseAutomatically = @"TmuxUnpauseAutomatically";
+NSString *const kPreferenceKeyTmuxWarnBeforePausing = @"TmuxWarnBeforePausing";
 
-NSString *const kPreferenceKeyTabStyle = @"TabStyle";
+NSString *const kPreferenceKeyUseMetal = @"UseMetal";
+NSString *const kPreferenceKeyDisableMetalWhenUnplugged = @"disableMetalWhenUnplugged";
+NSString *const kPreferenceKeyPreferIntegratedGPU = @"preferIntegratedGPU";
+NSString *const kPreferenceKeyMetalMaximizeThroughput = @"metalMaximizeThroughput";
+NSString *const kPreferenceKeyEnableAPIServer = @"EnableAPIServer";
+NSString *const kPreferenceKeyAPIAuthentication = @"API Authentication Method";
+
+NSString *const kPreferenceKeyTabStyle_Deprecated = @"TabStyle";  // Pre-10.14
+NSString *const kPreferenceKeyTabStyle = @"TabStyleWithAutomaticOption";  // Pre-10.14
 NSString *const kPreferenceKeyTabPosition = @"TabViewType";
+NSString *const kPreferenceKeyStatusBarPosition = @"StatusBarPosition";
 NSString *const kPreferenceKeyHideTabBar = @"HideTab";
 NSString *const kPreferenceKeyHideTabNumber = @"HideTabNumber";
-NSString *const kPreferenceKeyHideTabCloseButton = @"HideTabCloseButton";
+NSString *const kPreferenceKeyPreserveWindowSizeWhenTabBarVisibilityChanges = @"PreserveWindowSizeWhenTabBarVisibilityChanges";
+NSString *const kPreferenceKeyHideTabCloseButton = @"HideTabCloseButton";  // Deprecated
+NSString *const kPreferenceKeyTabsHaveCloseButton = @"TabsHaveCloseButton";
 NSString *const kPreferenceKeyHideTabActivityIndicator = @"HideActivityIndicator";
 NSString *const kPreferenceKeyShowNewOutputIndicator = @"ShowNewOutputIndicator";
 NSString *const kPreferenceKeyShowPaneTitles = @"ShowPaneTitles";
+NSString *const kPreferenceKeyPerPaneBackgroundImage = @"PerPaneBackgroundImage";
+NSString *const kPreferenceKeyStretchTabsToFillBar = @"StretchTabsToFillBar";
 NSString *const kPreferenceKeyHideMenuBarInFullscreen = @"HideMenuBarInFullscreen";
+NSString *const kPreferenceKeyUIElement = @"HideFromDockAndAppSwitcher";
+NSString *const kPreferenceKeyUIElementRequiresHotkeys = @"UIElementRequiresHotkeys";
 NSString *const kPreferenceKeyFlashTabBarInFullscreen = @"FlashTabBarInFullscreen";
 NSString *const kPreferenceKeyShowWindowNumber = @"WindowNumber";
-NSString *const kPreferenceKeyShowJobName = @"JobName";
-NSString *const kPreferenceKeyShowProfileName = @"ShowBookmarkName";  // The key predates bookmarks being renamed to profiles
+NSString *const kPreferenceKeyShowJobName_Deprecated = @"JobName";
+NSString *const kPreferenceKeyShowProfileName_Deprecated = @"ShowBookmarkName";  // The key predates bookmarks being renamed to profiles
 NSString *const kPreferenceKeyDimOnlyText = @"DimOnlyText";
 NSString *const kPreferenceKeyDimmingAmount = @"SplitPaneDimmingAmount";
 NSString *const kPreferenceKeyDimInactiveSplitPanes = @"DimInactiveSplitPanes";
 NSString *const kPreferenceKeyShowWindowBorder = @"UseBorder";
 NSString *const kPreferenceKeyHideScrollbar = @"HideScrollbar";
 NSString *const kPreferenceKeyDisableFullscreenTransparencyByDefault = @"DisableFullscreenTransparency";
+NSString *const kPreferenceKeyEnableDivisionView = @"EnableDivisionView";
+NSString *const kPreferenceKeyEnableProxyIcon = @"EnableProxyIcon";
 NSString *const kPreferenceKeyDimBackgroundWindows = @"DimBackgroundWindows";
+NSString *const kPreferenceKeySeparateStatusBarsPerPane = @"SeparateStatusBarsPerPane";
 
 NSString *const kPreferenceKeyControlRemapping = @"Control";
 NSString *const kPreferenceKeyLeftOptionRemapping = @"LeftOption";
@@ -82,13 +113,19 @@ NSString *const kPreferenceKeyRightCommandRemapping = @"RightCommand";
 NSString *const kPreferenceKeySwitchPaneModifier = @"SwitchPaneModifier";
 NSString *const kPreferenceKeySwitchTabModifier = @"SwitchTabModifier";
 NSString *const kPreferenceKeySwitchWindowModifier = @"SwitchWindowModifier";
+NSString *const kPreferenceKeyEmulateUSKeyboard = @"UseVirtualKeyCodesForDetectingDigits";
+
 NSString *const kPreferenceKeyHotkeyEnabled = @"Hotkey";
 NSString *const kPreferenceKeyHotKeyCode = @"HotkeyCode";
 NSString *const kPreferenceKeyHotkeyCharacter = @"HotkeyChar";  // Nonzero if hotkey char is set.
 NSString *const kPreferenceKeyHotkeyModifiers = @"HotkeyModifiers";
-NSString *const kPreferenceKeyHotKeyTogglesWindow = @"HotKeyTogglesWindow";
-NSString *const kPreferenceKeyHotkeyProfileGuid = @"HotKeyBookmark";
-NSString *const kPreferenceKeyHotkeyAutoHides = @"HotkeyAutoHides";
+NSString *const kPreferenceKeyEnableHapticFeedbackForEsc = @"HapticFeedbackForEsc";
+NSString *const kPreferenceKeyEnableSoundForEsc = @"SoundForEsc";
+NSString *const kPreferenceKeyVisualIndicatorForEsc = @"VisualIndicatorForEsc";
+
+NSString *const kPreferenceKeyHotKeyTogglesWindow_Deprecated = @"HotKeyTogglesWindow";  // deprecated
+NSString *const kPreferenceKeyHotkeyProfileGuid_Deprecated = @"HotKeyBookmark";  // deprecated
+NSString *const kPreferenceKeyHotkeyAutoHides_Deprecated = @"HotkeyAutoHides";  // deprecated
 
 NSString *const kPreferenceKeyCmdClickOpensURLs = @"CommandSelection";
 NSString *const kPreferenceKeyControlLeftClickBypassesContextMenu = @"PassOnControlClick";
@@ -97,9 +134,13 @@ NSString *const kPreferenceKeyThreeFingerEmulatesMiddle = @"ThreeFingerEmulates"
 NSString *const kPreferenceKeyFocusFollowsMouse = @"FocusFollowsMouse";
 NSString *const kPreferenceKeyTripleClickSelectsFullWrappedLines = @"TripleClickSelectsFullWrappedLines";
 NSString *const kPreferenceKeyDoubleClickPerformsSmartSelection = @"DoubleClickPerformsSmartSelection";
+NSString *const kPreferenceKeyEnterCopyModeAutomatically = @"EnterCopyModeAutomatically";
+NSString *const kPreferenceKeyFocusOnRightOrMiddleClick = @"FocusOnRightOrMiddleClick";
 
-NSString *const kPreferenceKeyAppVersion = @"iTerm Version";
+NSString *const kPreferenceKeyAppVersion = @"iTerm Version";  // Excluded from syncing
+NSString *const kPreferenceKeyAllAppVersions = @"NoSyncAllAppVersions";  // Array of known iTerm2 versions this user has used on this machine.
 NSString *const kPreferenceAutoCommandHistory = @"AutoCommandHistory";
+NSString *const kPreferenceKeyOSVersion = @"NoSyncLastOSVersion";
 
 NSString *const kPreferenceKeyPasteSpecialChunkSize = @"PasteSpecialChunkSize";
 NSString *const kPreferenceKeyPasteSpecialChunkDelay = @"PasteSpecialChunkDelay";
@@ -118,61 +159,140 @@ NSString *const kPreferenceKeyLeftTabBarWidth = @"LeftTabBarWidth";
 NSString *const kPreferenceKeyPasteWarningNumberOfSpacesPerTab = @"PasteTabToStringTabStopSize";
 
 NSString *const kPreferenceKeyShowFullscreenTabBar = @"ShowFullScreenTabBar";
+NSString *const kPreferenceKeyHotkeyMigratedFromSingleToMulti = @"HotkeyMigratedFromSingleToMulti";
+NSString *const kPreferenceKeyDefaultToolbeltWidth = @"Default Toolbelt Width";
+NSString *const kPreferenceKeySizeChangesAffectProfile = @"Size Changes Affect Profile";
+NSString *const kPreferenceKeyActions = @"Actions";
+NSString *const kPreferenceKeySnippets = @"Snippets";
+// NOTE: If you update this list, also update preferences.py.
 
 static NSMutableDictionary *gObservers;
+static NSString *sPreviousVersion;
 
 @implementation iTermPreferences
 
++ (NSString *)appVersionBeforeThisLaunch {
+    if (!sPreviousVersion) {
+        NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
+        sPreviousVersion = [[userDefaults objectForKey:kPreferenceKeyAppVersion] copy];
+    }
+    return sPreviousVersion;
+}
+
++ (NSSet<NSString *> *)allAppVersionsUsedOnThisMachine {
+    static NSSet<NSString *> *versions;
+    static dispatch_once_t onceToken;
+    dispatch_once(&onceToken, ^{
+        versions = [NSSet setWithArray:[[NSUserDefaults standardUserDefaults] objectForKey:kPreferenceKeyAllAppVersions] ?: @[]];
+    });
+    return versions;
+
+}
+
++ (void)initializeAppVersionBeforeThisLaunch:(NSString *)thisVersion {
+    // Force it to be lazy-loaded.
+    [self appVersionBeforeThisLaunch];
+    // Then overwrite it with the current version
+    [[NSUserDefaults standardUserDefaults] setObject:thisVersion forKey:kPreferenceKeyAppVersion];
+    NSProcessInfo *processInfo = [NSProcessInfo processInfo];
+    [[NSUserDefaults standardUserDefaults] setObject:processInfo.operatingSystemVersionString
+                                              forKey:kPreferenceKeyOSVersion];
+}
+
++ (void)initializeAllAppVersionsUsedOnThisMachine:(NSString *)thisVersion {
+    // Update all app versions ever seena.
+    NSMutableSet *allVersions = [NSMutableSet setWithArray:[[NSUserDefaults standardUserDefaults] objectForKey:kPreferenceKeyAllAppVersions] ?: @[]];
+    
+    NSString *const before = [self appVersionBeforeThisLaunch];
+    if (before) {
+        [allVersions addObject:before];
+    }
+
+    [allVersions addObject:thisVersion];
+    [allVersions removeObject:@"unknown"];
+    [[NSUserDefaults standardUserDefaults] setObject:allVersions.allObjects forKey:kPreferenceKeyAllAppVersions];
+}
+
++ (NSDictionary *)systemPreferenceOverrides {
+    return @{
+        // Force antialiasing to be allowed on small font sizes
+        @"AppleAntiAliasingThreshold": @1,
+        @"AppleSmoothFixedFontsSizeThreshold": @1,
+
+        // Turn off high sierra's native tabs
+        @"AppleWindowTabbingMode": @"manual",
+
+        // Turn off scroll animations because they screw up the terminal scrolling.
+        @"AppleScrollAnimationEnabled": @0,
+
+        // Turn off accent menu
+        @"ApplePressAndHoldEnabled": @NO,
+
+        // Override smooth scrolling, which breaks various things (such as the
+        // assumption, when detectUserScroll is called, that scrolls happen
+        // immediately), and generally sucks with a terminal.
+        @"NSScrollAnimationEnabled": @NO,
+
+        // Disable under-titlebar mirror view.
+
+        // OS 10.10 has a spiffy feature where it finds a scrollview that is
+        // adjacent to the title bar and then does some magic to makes the
+        // scrollview's content show up with "vibrancy" (i.e., blur) under the
+        // title bar. The way it does this is to create an "NSScrollViewMirrorView"
+        // in the title bar's view hierarchy, under a view whose class is
+        // NSTitlebarContainerView. Unfortunately there is no way to turn
+        // this off. You can move the scroll view at least two points away from the
+        // title bar, but that looks terrible. Terminal.app went so far as to stop
+        // using scroll views! Trying to replace NSScrollView with my custom
+        // implementation seems fraught with peril. Trying to hide the mirror view
+        // doesn't work because it only becomes visible once the scroll view is
+        // taller than the window's content view (I think that is new in 10.10.3).
+        // I tried swizzling addSubview: in NSTitlebarContainerView to hide
+        // mirror views when they get added, but it caused some performance problems
+        // I can't reproduce (see issue 3499).
+        //
+        // Another option, which seems more fragile, is to override
+        // -[PTYScrollView _makeUnderTitlebarView] and have it return nil. That works
+        // in testing but could break things pretty badly.
+        //
+        // I found this undocumented setting while disassembling the caller to _makeUnderTitlebarView,
+        // and it seems to work.
+        //
+        // See issue 3244 for details.
+        @"NSScrollViewShouldScrollUnderTitlebar": @NO,
+
+        // macOS does an insane thing and looks for views that might possibly be trying to do something
+        // clever with scrollers and then if it finds them changes to legacy scrollers. Quothe the
+        // mailing list https://lists.apple.com/archives/cocoa-dev/2012/Mar/msg00939.html
+        //  NSScrollView does various checks to see if the App is trying to put placards in the
+        //  scroller area. If NSScrollView thinks there are placards, then it reverts back to legacy
+        //  scrollers for compatibility. Some apps have been known to do this via a sibling view
+        //  instead of a subview. This is why it intermittently happens during your animation as
+        //  your sibling views momentarily overlap
+        //
+        // In addition to being a bad idea, it is poorly implemented and the scrollers randomly
+        // have the wrong light/dark appearance, in addition to other visual glitches that words
+        // can't describe.
+        //
+        // Well, I am trying to do something clever and I damn well want it to work. With a bit of
+        // the old disassembler, I found this user default which seems to turn off the stupid.
+        @"NSOverlayScrollersFallBackForAccessoryViews": @NO,
+    };
+}
+
 + (void)initializeUserDefaults {
     NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
+    [[self systemPreferenceOverrides] enumerateKeysAndObjectsUsingBlock:^(id  _Nonnull key, id  _Nonnull obj, BOOL * _Nonnull stop) {
+        [userDefaults setObject:obj forKey:key];
+    }];
 
-    // Force antialiasing to be allowed on small font sizes
-    [userDefaults setInteger:1 forKey:@"AppleAntiAliasingThreshold"];
-    [userDefaults setInteger:1 forKey:@"AppleSmoothFixedFontsSizeThreshold"];
-    
-    // Turn off scroll animations because they screw up the terminal scrolling.
-    [userDefaults setInteger:0 forKey:@"AppleScrollAnimationEnabled"];
-
-    // Override smooth scrolling, which breaks various things (such as the
-    // assumption, when detectUserScroll is called, that scrolls happen
-    // immediately), and generally sucks with a terminal.
-    [[NSUserDefaults standardUserDefaults] setBool:NO forKey:@"NSScrollAnimationEnabled"];
-
-    // Store the current app version in prefs
     NSDictionary *infoDictionary = [[NSBundle bundleForClass:[self class]] infoDictionary];
-    [userDefaults setObject:infoDictionary[@"CFBundleVersion"] forKey:kPreferenceKeyAppVersion];
-
-    // Disable under-titlebar mirror view.
-
-    // OS 10.10 has a spiffy feature where it finds a scrollview that is
-    // adjacent to the title bar and then does some magic to makes the
-    // scrollview's content show up with "vibrancy" (i.e., blur) under the
-    // title bar. The way it does this is to create an "NSScrollViewMirrorView"
-    // in the title bar's view hierarchy, under a view whose class is
-    // NSTitlebarContainerView. Unfortunately there is no way to turn
-    // this off. You can move the scroll view at least two points away from the
-    // title bar, but that looks terrible. Terminal.app went so far as to stop
-    // using scroll views! Trying to replace NSScrollView with my custom
-    // implementation seems fraught with peril. Trying to hide the mirror view
-    // doesn't work because it only becomes visible once the scroll view is
-    // taller than the window's content view (I think that is new in 10.10.3).
-    // I tried swizzling addSubview: in NSTitlebarContainerView to hide
-    // mirror views when they get added, but it caused some performance problems
-    // I can't reproduce (see issue 3499).
-    //
-    // Another option, which seems more fragile, is to override
-    // -[PTYScrollView _makeUnderTitlebarView] and have it return nil. That works
-    // in testing but could break things pretty badly.
-    //
-    // I found this undocumented setting while disassembling the caller to _makeUnderTitlebarView,
-    // and it seems to work.
-    //
-    // See issue 3244 for details.
-    [[NSUserDefaults standardUserDefaults] setBool:NO
-                                            forKey:@"NSScrollViewShouldScrollUnderTitlebar"];
+    NSString *const thisVersion = infoDictionary[@"CFBundleVersion"];
+    [self initializeAppVersionBeforeThisLaunch:thisVersion];
+    [self initializeAllAppVersionsUsedOnThisMachine:thisVersion];
 
     // Load prefs from remote.
-    [[iTermRemotePreferences sharedInstance] copyRemotePrefsToLocalUserDefaults];
+    [[iTermRemotePreferences sharedInstance] copyRemotePrefsToLocalUserDefaultsPreserving:self.systemPreferenceOverrides.allKeys];
 }
 
 #pragma mark - Default values
@@ -186,6 +306,7 @@ static NSMutableDictionary *gObservers;
                   kPreferenceKeyQuitWhenAllWindowsClosed: @NO,
                   kPreferenceKeyConfirmClosingMultipleTabs: @YES,
                   kPreferenceKeyPromptOnQuit: @YES,
+                  kPreferenceKeyPromptOnQuitEvenIfThereAreNoWindows: @NO,
                   kPreferenceKeyInstantReplayMemoryMegabytes: @4,
                   kPreferenceKeySavePasteAndCommandHistory: @NO,
                   kPreferenceKeyAddBonjourHostsToProfiles: @NO,
@@ -202,32 +323,56 @@ static NSMutableDictionary *gObservers;
                   kPreferenceKeySmartWindowPlacement: @NO,
                   kPreferenceKeyAdjustWindowForFontSizeChange: @YES,
                   kPreferenceKeyMaximizeVerticallyOnly: @NO,
-                  kPreferenceKeyLionStyleFullscren: @YES,
+                  kPreferenceKeyLionStyleFullscreen: @YES,
+                  kPreferenceKeySeparateWindowTitlePerTab: @YES,
                   kPreferenceKeyOpenTmuxWindowsIn: @(kOpenTmuxWindowsAsNativeWindows),
                   kPreferenceKeyTmuxDashboardLimit: @10,
                   kPreferenceKeyAutoHideTmuxClientSession: @NO,
+                  kPreferenceKeyUseTmuxProfile: @YES,
+                  kPreferenceKeyUseTmuxStatusBar: @YES,
+                  kPreferenceKeyTmuxPauseModeAgeLimit: @120,
+                  kPreferenceKeyTmuxUnpauseAutomatically: @NO,
+                  kPreferenceKeyTmuxWarnBeforePausing: @YES,
+                  kPreferenceKeyUseMetal: @YES,
+                  kPreferenceKeyDisableMetalWhenUnplugged: @YES,
+                  kPreferenceKeyPreferIntegratedGPU: @YES,
+                  kPreferenceKeyMetalMaximizeThroughput: @YES,
+                  kPreferenceKeyEnableAPIServer: @NO,
+                  kPreferenceKeyAPIAuthentication: @0,  // ignored — synthetic value
 
+                  kPreferenceKeyTabStyle_Deprecated: @(TAB_STYLE_LIGHT),
                   kPreferenceKeyTabStyle: @(TAB_STYLE_LIGHT),
+                  
                   kPreferenceKeyTabPosition: @(TAB_POSITION_TOP),
+                  kPreferenceKeyStatusBarPosition: @(iTermStatusBarPositionTop),
                   kPreferenceKeyHideTabBar: @YES,
                   kPreferenceKeyHideTabNumber: @NO,
-                  kPreferenceKeyHideTabCloseButton: @NO,
+                  kPreferenceKeyPreserveWindowSizeWhenTabBarVisibilityChanges: @NO,
+                  kPreferenceKeyHideTabCloseButton: @NO,  // Deprecated
+                  kPreferenceKeyTabsHaveCloseButton: @YES,
                   kPreferenceKeyHideTabActivityIndicator: @NO,
                   kPreferenceKeyShowNewOutputIndicator: @YES,
+                  kPreferenceKeyStretchTabsToFillBar: @YES,
 
                   kPreferenceKeyShowPaneTitles: @YES,
+                  kPreferenceKeyPerPaneBackgroundImage: @YES,
                   kPreferenceKeyHideMenuBarInFullscreen:@YES,
-                  kPreferenceKeyFlashTabBarInFullscreen:@YES,
+                  kPreferenceKeyUIElement: @NO,
+                  kPreferenceKeyUIElementRequiresHotkeys: @NO,
+                  kPreferenceKeyFlashTabBarInFullscreen:@NO,
                   kPreferenceKeyShowWindowNumber: @YES,
-                  kPreferenceKeyShowJobName: @YES,
-                  kPreferenceKeyShowProfileName: @NO,
+                  kPreferenceKeyShowJobName_Deprecated: @YES,
+                  kPreferenceKeyShowProfileName_Deprecated: @NO,
                   kPreferenceKeyDimOnlyText: @NO,
                   kPreferenceKeyDimmingAmount: @0.4,
                   kPreferenceKeyDimInactiveSplitPanes: @YES,
                   kPreferenceKeyShowWindowBorder: @NO,
                   kPreferenceKeyHideScrollbar: @NO,
                   kPreferenceKeyDisableFullscreenTransparencyByDefault: @NO,
+                  kPreferenceKeyEnableDivisionView: @YES,
+                  kPreferenceKeyEnableProxyIcon: @NO,
                   kPreferenceKeyDimBackgroundWindows: @NO,
+                  kPreferenceKeySeparateStatusBarsPerPane: @NO,
 
                   kPreferenceKeyControlRemapping: @(kPreferencesModifierTagControl),
                   kPreferenceKeyLeftOptionRemapping: @(kPreferencesModifierTagLeftOption),
@@ -237,21 +382,27 @@ static NSMutableDictionary *gObservers;
                   kPreferenceKeySwitchPaneModifier: @(kPreferenceModifierTagNone),
                   kPreferenceKeySwitchTabModifier: @(kPreferencesModifierTagEitherCommand),
                   kPreferenceKeySwitchWindowModifier: @(kPreferencesModifierTagCommandAndOption),
+                  kPreferenceKeyEmulateUSKeyboard: @NO,
                   kPreferenceKeyHotkeyEnabled: @NO,
                   kPreferenceKeyHotKeyCode: @0,
                   kPreferenceKeyHotkeyCharacter: @0,
                   kPreferenceKeyHotkeyModifiers: @0,
-                  kPreferenceKeyHotKeyTogglesWindow: @NO,
-                  kPreferenceKeyHotkeyProfileGuid: [NSNull null],
-                  kPreferenceKeyHotkeyAutoHides: @YES,
+                  kPreferenceKeyHotKeyTogglesWindow_Deprecated: @NO,
+                  kPreferenceKeyHotkeyProfileGuid_Deprecated: [NSNull null],
+                  kPreferenceKeyHotkeyAutoHides_Deprecated: @YES,
+                  kPreferenceKeyEnableHapticFeedbackForEsc: @NO,
+                  kPreferenceKeyEnableSoundForEsc: @NO,
+                  kPreferenceKeyVisualIndicatorForEsc: @NO,
 
                   kPreferenceKeyCmdClickOpensURLs: @YES,
                   kPreferenceKeyControlLeftClickBypassesContextMenu: @NO,
                   kPreferenceKeyOptionClickMovesCursor: @YES,
                   kPreferenceKeyThreeFingerEmulatesMiddle: @NO,
                   kPreferenceKeyFocusFollowsMouse: @NO,
-                  kPreferenceKeyTripleClickSelectsFullWrappedLines: @NO,
+                  kPreferenceKeyFocusOnRightOrMiddleClick: @NO,
+                  kPreferenceKeyTripleClickSelectsFullWrappedLines: @YES,
                   kPreferenceKeyDoubleClickPerformsSmartSelection: @NO,
+                  kPreferenceKeyEnterCopyModeAutomatically: @YES,
 
                   kPreferenceAutoCommandHistory: @NO,
 
@@ -270,9 +421,11 @@ static NSMutableDictionary *gObservers;
 
                   kPreferenceKeyPasteWarningNumberOfSpacesPerTab: @4,
                   kPreferenceKeyShowFullscreenTabBar: @YES,
+                  kPreferenceKeyHotkeyMigratedFromSingleToMulti: @NO,
                   kPreferenceKeyLeftTabBarWidth: @150,
+                  kPreferenceKeyDefaultToolbeltWidth: @250,
+                  kPreferenceKeySizeChangesAffectProfile: @NO,
               };
-        [dict retain];
     }
     return dict;
 }
@@ -312,10 +465,12 @@ static NSMutableDictionary *gObservers;
                     [defaultValue isKindOfClass:[NSNull class]]);
         case kPreferenceInfoTypeMatrix:
             return [defaultValue isKindOfClass:[NSString class]];
+        case kPreferenceInfoTypeRadioButton:
+            return [defaultValue isKindOfClass:[NSString class]];
         case kPreferenceInfoTypeColorWell:
             return [defaultValue isKindOfClass:[NSDictionary class]];
     }
-    
+
     return NO;
 }
 
@@ -329,14 +484,17 @@ static NSMutableDictionary *gObservers;
     if (!dict) {
         dict = @{ kPreferenceKeyOpenArrangementAtStartup: BLOCK(computedOpenArrangementAtStartup),
                   kPreferenceKeyCustomFolder: BLOCK(computedCustomFolder),
-                  kPreferenceKeyCharactersConsideredPartOfAWordForSelection: BLOCK(computedWordChars) };
-        [dict retain];
+                  kPreferenceKeyCharactersConsideredPartOfAWordForSelection: BLOCK(computedWordChars),
+                  kPreferenceKeyTabStyle: BLOCK(computedTabStyle),
+                  kPreferenceKeyUseMetal: BLOCK(computedUseMetal),
+                  kPreferenceKeyTabsHaveCloseButton: BLOCK(computedTabsHaveCloseButton)
+                  };
     }
     return dict;
 }
 
 + (id)computedObjectForKey:(NSString *)key {
-    id (^block)() = [self computedObjectDictionary][key];
+    id (^block)(void) = [self computedObjectDictionary][key];
     if (block) {
         return block();
     } else {
@@ -365,7 +523,7 @@ static NSMutableDictionary *gObservers;
     id before = nil;
     if (observers) {
         before = [self objectForKey:key];
-        
+
         // nil out observers if there is no change.
         if (before && object && [before isEqual:object]) {
             observers = nil;
@@ -382,6 +540,8 @@ static NSMutableDictionary *gObservers;
     for (void (^block)(id, id) in observers) {
         block(before, object);
     }
+
+    [[iTermPreferenceDidChangeNotification notificationWithKey:key value:object] post];
 }
 
 #pragma mark - APIs
@@ -406,6 +566,14 @@ static NSMutableDictionary *gObservers;
     [self setObject:@(value) forKey:key];
 }
 
++ (NSInteger)integerForKey:(NSString *)key {
+    return [(NSNumber *)[self objectForKey:key] integerValue];
+}
+
++ (void)setInteger:(NSInteger)value forKey:(NSString *)key {
+    [self setObject:@(value) forKey:key];
+}
+
 + (NSUInteger)unsignedIntegerForKey:(NSString *)key {
     return [(NSNumber *)[self objectForKey:key] unsignedIntegerValue];
 }
@@ -449,26 +617,29 @@ static NSMutableDictionary *gObservers;
         observersForKey = [NSMutableArray array];
         gObservers[key] = observersForKey;
     }
-    [observersForKey addObject:[[block copy] autorelease]];
+    [observersForKey addObject:[block copy]];
 }
 
 + (NSUInteger)maskForModifierTag:(iTermPreferencesModifierTag)tag {
     switch (tag) {
+        case kPreferencesModifierTagControl:
+            return NSEventModifierFlagControl;
+
         case kPreferencesModifierTagEitherCommand:
-            return NSCommandKeyMask;
-            
+            return NSEventModifierFlagCommand;
+
         case kPreferencesModifierTagCommandAndOption:
-            return NSCommandKeyMask | NSAlternateKeyMask;
-            
+            return NSEventModifierFlagCommand | NSEventModifierFlagOption;
+
         case kPreferencesModifierTagEitherOption:
-            return NSAlternateKeyMask;
+            return NSEventModifierFlagOption;
 
         case kPreferenceModifierTagNone:
             return NSUIntegerMax;
 
         default:
             NSLog(@"Unexpected value for maskForModifierTag: %d", tag);
-            return NSCommandKeyMask | NSAlternateKeyMask;
+            return NSEventModifierFlagCommand | NSEventModifierFlagOption;
     }
 }
 
@@ -495,4 +666,96 @@ static NSMutableDictionary *gObservers;
     return wordChars ?: @"";
 }
 
+// Migrates all pre-10.14 users now on 10.14 to automatic, since anything else looks bad.
++ (NSNumber *)computedTabStyle {
+    NSNumber *value;
+    value = [[NSUserDefaults standardUserDefaults] objectForKey:kPreferenceKeyTabStyle];
+    if (value) {
+        return value;
+    }
+    if (@available(macOS 10.14, *)) {
+        // New value is not set yet. This migrates all users to automatic.
+        return @(TAB_STYLE_AUTOMATIC);
+    }
+    value = [[NSUserDefaults standardUserDefaults] objectForKey:kPreferenceKeyTabStyle_Deprecated];
+    if (value) {
+        return value;
+    } else {
+        return @(TAB_STYLE_LIGHT);
+    }
+}
+
++ (NSNumber *)computedTabsHaveCloseButton {
+    NSNumber *value;
+    value = [[NSUserDefaults standardUserDefaults] objectForKey:kPreferenceKeyTabsHaveCloseButton];
+    if (value) {
+        return value;
+    }
+
+    value = [[NSUserDefaults standardUserDefaults] objectForKey:@"eliminateCloseButtons"];
+    if (value) {
+        return @(!value.boolValue);
+    }
+
+    return [self defaultObjectForKey:kPreferenceKeyTabsHaveCloseButton];
+}
+
++ (NSNumber *)computedUseMetal {
+    NSNumber *value;
+    value = [[NSUserDefaults standardUserDefaults] objectForKey:kPreferenceKeyUseMetal];
+    if (value) {
+        return value;
+    }
+
+    if (@available(macOS 10.13, *)) {
+        return @YES;
+    }
+
+    // Off by default on 10.12 because it's slow.
+    return @NO;
+}
+
++ (iTermUserDefaultsObserver *)sharedObserver {
+    static iTermUserDefaultsObserver *instance;
+    static dispatch_once_t onceToken;
+    dispatch_once(&onceToken, ^{
+        instance = [[iTermUserDefaultsObserver alloc] init];
+    });
+    return instance;
+}
+
+@end
+
+typedef struct {
+    NSString *key;
+    BOOL value;
+    dispatch_once_t onceToken;
+} iTermPreferencesBoolCache;
+
+#define FAST_BOOL_ACCESSOR(accessorName, userDefaultsKey) \
++ (BOOL)accessorName { \
+    static iTermPreferencesBoolCache cache = { \
+        .key = userDefaultsKey, \
+        .value = NO, \
+        .onceToken = 0 \
+    }; \
+    return [self boolWithCache:&cache]; \
+}
+
+@implementation iTermPreferences (FastAccessors)
+
++ (BOOL)boolWithCache:(iTermPreferencesBoolCache *)cache {
+    dispatch_once(&cache->onceToken, ^{
+        cache->value = [self boolForKey:cache->key];
+        [[self sharedObserver] observeKey:cache->key block:^{
+            cache->value = [self boolForKey:cache->key];
+        }];
+    });
+    return cache->value;
+}
+
+FAST_BOOL_ACCESSOR(hideTabActivityIndicator, kPreferenceKeyHideTabActivityIndicator)
+FAST_BOOL_ACCESSOR(maximizeMetalThroughput, kPreferenceKeyMetalMaximizeThroughput)
+FAST_BOOL_ACCESSOR(useTmuxProfile, kPreferenceKeyUseTmuxProfile)
+
 @end
