@@ -9,7 +9,6 @@
 
 #import "DebugLogging.h"
 #import "iTermStatusBarAutoRainbowController.h"
-#import "iTermStatusBarContainerView.h"
 #import "iTermStatusBarFixedSpacerComponent.h"
 #import "iTermStatusBarLayout.h"
 #import "iTermStatusBarLayoutAlgorithm.h"
@@ -28,9 +27,6 @@
 NS_ASSUME_NONNULL_BEGIN
 
 static const CGFloat iTermStatusBarViewControllerBottomMargin = 0;
-static const CGFloat iTermStatusBarViewControllerContainerHeight = 21;
-
-const CGFloat iTermStatusBarHeight = 21;
 
 @interface iTermStatusBarViewController ()<
     iTermStatusBarComponentDelegate,
@@ -91,7 +87,8 @@ const CGFloat iTermStatusBarHeight = 21;
     return [iTermStatusBarLayoutAlgorithm layoutAlgorithmWithContainerViews:_containerViews
                                                               mandatoryView:self.mandatoryView
                                                              statusBarWidth:self.view.frame.size.width
-                                                                    setting:_layout.advancedConfiguration.layoutAlgorithm];
+                                                                    setting:_layout.advancedConfiguration.layoutAlgorithm
+                                                      removeEmptyComponents:_layout.advancedConfiguration.removeEmptyComponents];
 }
 
 - (NSArray<NSNumber *> *)desiredSeparatorOffsets {
@@ -161,7 +158,7 @@ const CGFloat iTermStatusBarHeight = 21;
          view.frame = NSMakeRect(round(view.desiredOrigin),
                                  iTermStatusBarViewControllerBottomMargin,
                                  ceil(view.desiredWidth),
-                                 iTermStatusBarViewControllerContainerHeight);
+                                 iTermGetStatusBarHeight() - iTermStatusBarViewControllerBottomMargin);
          [view.component statusBarComponentWidthDidChangeTo:view.desiredWidth];
          [view layoutSubviews];
          view.rightSeparatorOffset = -1;
@@ -302,6 +299,12 @@ const CGFloat iTermStatusBarHeight = 21;
     }].component;
 }
 
+- (nullable __kindof id<iTermStatusBarComponent>)visibleComponentWithIdentifier:(NSString *)identifier {
+    return [_visibleContainerViews objectPassingTest:^BOOL(iTermStatusBarContainerView *element, NSUInteger index, BOOL *stop) {
+        return [element.component.statusBarComponentIdentifier isEqual:identifier];
+    }].component;
+}
+
 #pragma mark - iTermStatusBarLayoutDelegate
 
 - (void)statusBarLayoutDidChange:(iTermStatusBarLayout *)layout {
@@ -355,10 +358,26 @@ const CGFloat iTermStatusBarHeight = 21;
     [self.delegate statusBarPerformAction:action];
 }
 
-- (void)statusBarComponentRevealActionsTool:(id<iTermStatusBarComponent>)component {
-    [self.delegate statusBarRevealActionsTool];
+- (void)statusBarComponentEditActions:(id<iTermStatusBarComponent>)component {
+    [self.delegate statusBarEditActions];
 }
 
+- (void)statusBarComponentEditSnippets:(id<iTermStatusBarComponent>)component {
+    [self.delegate statusBarEditSnippets];
+}
+
+- (void)statusBarComponentResignFirstResponder:(id<iTermStatusBarComponent>)component {
+    [self.delegate statusBarResignFirstResponder];
+}
+
+- (void)statusBarComponent:(id<iTermStatusBarComponent>)component
+      reportScriptingError:(NSError *)error
+             forInvocation:(NSString *)invocation
+                    origin:(NSString *)origin {
+    [self.delegate statusBarReportScriptingError:error forInvocation:invocation origin:origin];
+}
+
+
 #pragma mark - iTermStatusBarContainerViewDelegate
 
 - (void)statusBarContainerView:(iTermStatusBarContainerView *)sender hideComponent:(id<iTermStatusBarComponent>)component {
@@ -413,7 +432,11 @@ const CGFloat iTermStatusBarHeight = 21;
     [_autoRainbowController enumerateColorsWithCount:_visibleContainerViews.count block:^(NSInteger i, NSColor * _Nonnull color) {
         id<iTermStatusBarComponent> component = _visibleContainerViews[i].component;
         NSMutableDictionary *knobValues = [component.configuration[iTermStatusBarComponentConfigurationKeyKnobValues] mutableCopy];
-        knobValues[iTermStatusBarSharedTextColorKey] = [color dictionaryValue];
+        NSDictionary *colorDict = [color dictionaryValue];
+        if ([knobValues[iTermStatusBarSharedTextColorKey] isEqualToDictionary:colorDict]) {
+            return;
+        }
+        knobValues[iTermStatusBarSharedTextColorKey] = colorDict;
         [component statusBarComponentSetKnobValues:knobValues];
     }];
 }
