@@ -1,4 +1,7 @@
 #!/bin/bash
+
+set -x
+
 function die {
   echo $1
   exit
@@ -9,12 +12,21 @@ if [ $# -ne 1 ]; then
    exit 1
 fi
 
-test -f "$PRIVKEY" || die "Set PRIVKEY environment variable to point at a valid private key (not set or nonexistent)"
+echo Enter the EdDSA private key
+read -s EDPRIVKEY
+
+echo Enter the notarization password
+read -s NOTPASS
+
 # Usage: SparkleSign testing.xml template.xml
 function SparkleSign {
     LENGTH=$(ls -l iTerm2-${NAME}.zip | awk '{print $5}')
-    ruby "../../ThirdParty/SparkleSigningTools/sign_update.rb" iTerm2-${NAME}.zip $PRIVKEY > /tmp/sig.txt || die SparkleSign
-    SIG=$(cat /tmp/sig.txt)
+
+    ../../tools/sign_update iTerm2-${NAME}.zip "$EDPRIVKEY" > /tmp/newsig.txt || die SparkleSignNew
+    echo "New signature is"
+    cat /tmp/newsig.txt
+
+    NEWSIG=$(cat /tmp/newsig.txt)
     DATE=$(date +"%a, %d %b %Y %H:%M:%S %z")
     XML=$1
     TEMPLATE=$2
@@ -24,8 +36,11 @@ function SparkleSign {
     sed -e "s/%VER%/${VERSION}/" | \
     sed -e "s/%DATE%/${DATE}/" | \
     sed -e "s/%NAME%/${NAME}/" | \
-    sed -e "s/%LENGTH%/$LENGTH/" |
-    sed -e "s,%SIG%,${SIG}," > $SVNDIR/source/appcasts/$1
+    sed -e "s/%LENGTH%/$LENGTH/" | \
+    sed -e "s,%NEWSIG%,${NEWSIG}," > $SVNDIR/source/appcasts/$1
+
+    echo "Updated appcasts file $SVNDIR/source/appcasts/$1"
+    cat $SVNDIR/source/appcasts/$1
     cp iTerm2-${NAME}.zip ~/iterm2-website/downloads/beta/
 }
 
@@ -52,8 +67,23 @@ function Build {
   rm -rf iTerm.app
   mv iTerm2.app iTerm.app
 
-  zip -ry iTerm2-${NAME}.zip iTerm.app
- 
+  # Zip it, notarize it, staple it, and re-zip it.
+  PRENOTARIZED_ZIP=iTerm2-${NAME}-prenotarized.zip
+  zip -ry $PRENOTARIZED_ZIP iTerm.app
+  xcrun altool --notarize-app --primary-bundle-id "com.googlecode.iterm2" --username "apple@georgester.com" --password "$NOTPASS" --file $PRENOTARIZED_ZIP > /tmp/upload.out 2>&1 || die "Notarization failed"
+  UUID=$(grep RequestUUID /tmp/upload.out | sed -e 's/RequestUUID = //')
+  echo "uuid is $UUID"
+  xcrun altool --notarization-info $UUID -u "apple@georgester.com" -p "$NOTPASS"
+  sleep 1
+  while xcrun altool --notarization-info $UUID -u "apple@georgester.com" -p "$NOTPASS" 2>&1 | egrep -i "in progress|Could not find the RequestUUID":
+  do
+      echo "Trying again"
+      sleep 1
+  done
+  NOTARIZED_ZIP=iTerm2-${NAME}.zip
+  xcrun stapler staple iTerm.app
+  zip -ry $NOTARIZED_ZIP iTerm.app
+
   # Update the list of changes
   vi $SVNDIR/source/appcasts/testing_changes3.txt
  
@@ -67,11 +97,17 @@ function Build {
   shasum -a256 iTerm2-${NAME}.zip | awk '{print $1}' >> $SVNDIR/downloads/beta/iTerm2-${NAME}.changelog
   vi $SVNDIR/downloads/beta/iTerm2-${NAME}.changelog
   pushd $SVNDIR
-  git add downloads/beta/iTerm2-${NAME}.summary downloads/beta/iTerm2-${NAME}.description downloads/beta/iTerm2-${NAME}.changelog downloads/beta/iTerm2-${NAME}.zip source/appcasts/testing3.xml source/appcasts/testing_changes3.txt
+  git add downloads/beta/iTerm2-${NAME}.summary downloads/beta/iTerm2-${NAME}.description downloads/beta/iTerm2-${NAME}.changelog downloads/beta/iTerm2-${NAME}.zip source/appcasts/testing3_new.xml source/appcasts/testing3_modern.xml source/appcasts/testing_changes3.txt
   popd
 
-  # Prepare the sparkle xml file
-  SparkleSign ${SPARKLE_PREFIX}testing3.xml ${SPARKLE_PREFIX}template3.xml
+  # Transitional
+  SparkleSign ${SPARKLE_PREFIX}testing3_new.xml ${SPARKLE_PREFIX}template3_new.xml
+  # Modern
+  SparkleSign ${SPARKLE_PREFIX}testing3_modern.xml ${SPARKLE_PREFIX}template3_modern.xml
+
+  # Copy experiment to control
+  cp $SVNDIR/source/appcasts/${SPARKLE_PREFIX}testing3_new.xml    $SVNDIR/source/appcasts/${SPARKLE_PREFIX}testing_new.xml
+  cp $SVNDIR/source/appcasts/${SPARKLE_PREFIX}testing3_modern.xml $SVNDIR/source/appcasts/${SPARKLE_PREFIX}testing_modern.xml
 
   popd
 }
@@ -88,11 +124,11 @@ ORIG_DIR=`pwd`
 
 echo "Build beta release"
 make clean
-make Beta
+make -j8 Beta
 
 BUILDTYPE=Beta
 
-Build $BUILDTYPE "" "OS 10.10+" "This is the recommended beta build for most users." "" "--deep"
+Build $BUILDTYPE "" "OS 10.14+" "This is the recommended beta build for most users." "" "--deep"
 
 git checkout -- version.txt
 #set -x
